import pytest
from prefect_gcp.credentials import GcpCredentials
from prefect_gcp.models.cloud_run_v2 import SecretKeySelector
from prefect_gcp.utilities import slugify_name
from prefect_gcp.workers.cloud_run_v2 import CloudRunWorkerJobV2Configuration

from prefect.utilities.dockerutils import get_prefect_image_name


@pytest.fixture
def job_body():
    return {
        "client": "prefect",
        "launchStage": None,
        "template": {
            "template": {
                "maxRetries": None,
                "timeout": None,
                "vpcAccess": {
                    "connector": None,
                },
                "containers": [
                    {
                        "env": [],
                        "command": None,
                        "args": "-m prefect.engine",
                        "resources": {
                            "limits": {
                                "cpu": None,
                                "memory": None,
                            },
                        },
                    },
                ],
            }
        },
    }


@pytest.fixture
def cloud_run_worker_v2_job_config(service_account_info, job_body):
    return CloudRunWorkerJobV2Configuration(
        name="my-job-name",
        job_body=job_body,
        credentials=GcpCredentials(service_account_info=service_account_info),
        region="us-central1",
        timeout=86400,
        env={"ENV1": "VALUE1", "ENV2": "VALUE2"},
    )


@pytest.fixture
def cloud_run_worker_v2_job_config_noncompliant_name(service_account_info, job_body):
    return CloudRunWorkerJobV2Configuration(
        name="MY_JOB_NAME",
        job_body=job_body,
        credentials=GcpCredentials(service_account_info=service_account_info),
        region="us-central1",
        timeout=86400,
        env={"ENV1": "VALUE1", "ENV2": "VALUE2"},
    )


class TestCloudRunWorkerJobV2Configuration:
    def test_project(self, cloud_run_worker_v2_job_config):
        assert cloud_run_worker_v2_job_config.project == "my_project"

    def test_job_name(self, cloud_run_worker_v2_job_config):
        assert cloud_run_worker_v2_job_config.job_name[:-33] == "my-job-name"

    def test_job_name_is_slug(self, cloud_run_worker_v2_job_config_noncompliant_name):
        assert cloud_run_worker_v2_job_config_noncompliant_name.job_name[
            :-33
        ] == slugify_name("MY_JOB_NAME")

    def test_job_name_different_after_retry(self, cloud_run_worker_v2_job_config):
        job_name_1 = cloud_run_worker_v2_job_config.job_name

        cloud_run_worker_v2_job_config._job_name = None

        job_name_2 = cloud_run_worker_v2_job_config.job_name

        assert job_name_1[:-33] == job_name_2[:-33]
        assert job_name_1 != job_name_2

    def test_populate_timeout(self, cloud_run_worker_v2_job_config):
        cloud_run_worker_v2_job_config._populate_timeout()

        assert (
            cloud_run_worker_v2_job_config.job_body["template"]["template"]["timeout"]
            == "86400s"
        )

    def test_populate_env(self, cloud_run_worker_v2_job_config):
        cloud_run_worker_v2_job_config._populate_env()

        assert cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"] == [
            {"name": "ENV1", "value": "VALUE1"},
            {"name": "ENV2", "value": "VALUE2"},
        ]

    def test_populate_env_with_secrets(self, cloud_run_worker_v2_job_config):
        cloud_run_worker_v2_job_config.env_from_secrets = {
            "SECRET_ENV1": SecretKeySelector(secret="SECRET1", version="latest")
        }
        cloud_run_worker_v2_job_config._populate_env()

        assert cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"] == [
            {"name": "ENV1", "value": "VALUE1"},
            {"name": "ENV2", "value": "VALUE2"},
            {
                "name": "SECRET_ENV1",
                "valueSource": {
                    "secretKeyRef": {"secret": "SECRET1", "version": "latest"}
                },
            },
        ]

    def test_populate_env_with_existing_envs(self, cloud_run_worker_v2_job_config):
        cloud_run_worker_v2_job_config.job_body["template"]["template"]["containers"][
            0
        ]["env"] = [{"name": "ENV0", "value": "VALUE0"}]
        cloud_run_worker_v2_job_config.env_from_secrets = {
            "SECRET_ENV1": SecretKeySelector(secret="SECRET1", version="latest")
        }
        cloud_run_worker_v2_job_config._populate_env()

        assert cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"] == [
            {"name": "ENV0", "value": "VALUE0"},
            {"name": "ENV1", "value": "VALUE1"},
            {"name": "ENV2", "value": "VALUE2"},
            {
                "name": "SECRET_ENV1",
                "valueSource": {
                    "secretKeyRef": {"secret": "SECRET1", "version": "latest"}
                },
            },
        ]

    def test_populate_image_if_not_present(self, cloud_run_worker_v2_job_config):
        cloud_run_worker_v2_job_config._populate_image_if_not_present()

        assert (
            cloud_run_worker_v2_job_config.job_body["template"]["template"][
                "containers"
            ][0]["image"]
            == f"docker.io/{get_prefect_image_name()}"
        )

    def test_populate_or_format_command(self, cloud_run_worker_v2_job_config):
        cloud_run_worker_v2_job_config._populate_or_format_command()

        assert cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["command"] == ["prefect", "flow-run", "execute"]

    def test_format_args_if_present(self, cloud_run_worker_v2_job_config):
        cloud_run_worker_v2_job_config._format_args_if_present()

        assert cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["args"] == ["-m", "prefect.engine"]

    @pytest.mark.parametrize("vpc_access", [{"connector": None}, {}, None])
    def test_remove_vpc_access_if_connector_unset(
        self, cloud_run_worker_v2_job_config, vpc_access
    ):
        cloud_run_worker_v2_job_config.job_body["template"]["template"]["vpcAccess"] = (
            vpc_access
        )

        cloud_run_worker_v2_job_config._remove_vpc_access_if_unset()

        assert (
            "vpcAccess"
            not in cloud_run_worker_v2_job_config.job_body["template"]["template"]
        )

    def test_remove_vpc_access_originally_not_present(
        self, cloud_run_worker_v2_job_config
    ):
        cloud_run_worker_v2_job_config.job_body["template"]["template"].pop("vpcAccess")

        cloud_run_worker_v2_job_config._remove_vpc_access_if_unset()

        assert (
            "vpcAccess"
            not in cloud_run_worker_v2_job_config.job_body["template"]["template"]
        )

    def test_vpc_access_left_alone_if_connector_set(
        self, cloud_run_worker_v2_job_config
    ):
        cloud_run_worker_v2_job_config.job_body["template"]["template"]["vpcAccess"][
            "connector"
        ] = "projects/my_project/locations/us-central1/connectors/my-connector"

        cloud_run_worker_v2_job_config._remove_vpc_access_if_unset()

        assert cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "vpcAccess"
        ] == {
            "connector": "projects/my_project/locations/us-central1/connectors/my-connector"  # noqa E501
        }

    def test_vpc_access_left_alone_if_network_config_set(
        self, cloud_run_worker_v2_job_config
    ):
        cloud_run_worker_v2_job_config.job_body["template"]["template"]["vpcAccess"][
            "networkInterfaces"
        ] = [{"network": "projects/my_project/global/networks/my-network"}]

        cloud_run_worker_v2_job_config._remove_vpc_access_if_unset()

        assert cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "vpcAccess"
        ] == {
            "connector": None,
            "networkInterfaces": [
                {"network": "projects/my_project/global/networks/my-network"}
            ],
        }

    def test_configure_cloudsql_volumes_no_instances(
        self, cloud_run_worker_v2_job_config
    ):
        cloud_run_worker_v2_job_config.cloudsql_instances = []
        cloud_run_worker_v2_job_config._configure_cloudsql_volumes()

        template = cloud_run_worker_v2_job_config.job_body["template"]["template"]

        assert "volumes" not in template
        assert "volumeMounts" not in template["containers"][0]

    def test_configure_cloudsql_volumes_preserves_existing_volumes(
        self, cloud_run_worker_v2_job_config
    ):
        template = cloud_run_worker_v2_job_config.job_body["template"]["template"]
        template["volumes"] = [{"name": "existing-volume", "emptyDir": {}}]
        template["containers"][0]["volumeMounts"] = [
            {"name": "existing-volume", "mountPath": "/existing"}
        ]

        cloud_run_worker_v2_job_config.cloudsql_instances = ["project:region:instance1"]
        cloud_run_worker_v2_job_config._configure_cloudsql_volumes()

        assert len(template["volumes"]) == 2
        assert template["volumes"][0] == {"name": "existing-volume", "emptyDir": {}}
        assert template["volumes"][1] == {
            "name": "cloudsql",
            "cloudSqlInstance": {"instances": ["project:region:instance1"]},
        }

        assert len(template["containers"][0]["volumeMounts"]) == 2
        assert template["containers"][0]["volumeMounts"][0] == {
            "name": "existing-volume",
            "mountPath": "/existing",
        }
        assert template["containers"][0]["volumeMounts"][1] == {
            "name": "cloudsql",
            "mountPath": "/cloudsql",
        }

    def test_prepare_for_flow_run_configures_cloudsql(
        self, cloud_run_worker_v2_job_config
    ):
        cloud_run_worker_v2_job_config.cloudsql_instances = ["project:region:instance1"]

        class MockFlowRun:
            id = "test-id"
            name = "test-run"

        cloud_run_worker_v2_job_config.prepare_for_flow_run(
            flow_run=MockFlowRun(), deployment=None, flow=None
        )

        template = cloud_run_worker_v2_job_config.job_body["template"]["template"]

        assert any(
            vol["name"] == "cloudsql"
            and vol["cloudSqlInstance"]["instances"] == ["project:region:instance1"]
            for vol in template["volumes"]
        )
        assert any(
            mount["name"] == "cloudsql" and mount["mountPath"] == "/cloudsql"
            for mount in template["containers"][0]["volumeMounts"]
        )

    def test_populate_env_with_prefect_api_key_secret(
        self, cloud_run_worker_v2_job_config
    ):
        cloud_run_worker_v2_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="prefect-api-key", version="latest"
        )
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars
        assert {
            "name": "PREFECT_API_KEY",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-api-key", "version": "latest"}
            },
        } in env_vars

    def test_populate_env_with_prefect_api_auth_string_secret(
        self, cloud_run_worker_v2_job_config
    ):
        cloud_run_worker_v2_job_config.prefect_api_auth_string_secret = (
            SecretKeySelector(secret="prefect-auth-string", version="latest")
        )
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-auth-string", "version": "latest"}
            },
        } in env_vars

    def test_populate_env_with_both_prefect_secrets(
        self, cloud_run_worker_v2_job_config
    ):
        cloud_run_worker_v2_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="prefect-api-key", version="latest"
        )
        cloud_run_worker_v2_job_config.prefect_api_auth_string_secret = (
            SecretKeySelector(secret="prefect-auth-string", version="latest")
        )
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars
        assert {
            "name": "PREFECT_API_KEY",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-api-key", "version": "latest"}
            },
        } in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-auth-string", "version": "latest"}
            },
        } in env_vars

    def test_populate_env_with_all_secret_types(self, cloud_run_worker_v2_job_config):
        cloud_run_worker_v2_job_config.env_from_secrets = {
            "SECRET_ENV1": SecretKeySelector(secret="SECRET1", version="latest")
        }
        cloud_run_worker_v2_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="prefect-api-key", version="latest"
        )
        cloud_run_worker_v2_job_config.prefect_api_auth_string_secret = (
            SecretKeySelector(secret="prefect-auth-string", version="latest")
        )
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars
        assert {
            "name": "SECRET_ENV1",
            "valueSource": {"secretKeyRef": {"secret": "SECRET1", "version": "latest"}},
        } in env_vars
        assert {
            "name": "PREFECT_API_KEY",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-api-key", "version": "latest"}
            },
        } in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-auth-string", "version": "latest"}
            },
        } in env_vars



import logging
import os
import tempfile
from pathlib import Path
from typing import Optional, cast

import typer
from pydantic_core import from_json

import prefect.runner
import prefect_gcp.credentials
from prefect.utilities.asyncutils import run_coro_as_sync

logger = logging.getLogger("prefect_gcp.experimental.bundles.execute")


def execute_bundle_from_gcs(
    bucket: str,
    key: str,
    gcp_credentials_block_name: Optional[str] = None,
) -> None:
    if isinstance(gcp_credentials_block_name, str):
        logger.debug(
            "Loading GCP credentials from block %s", gcp_credentials_block_name
        )
        gcp_credentials = cast(
            prefect_gcp.credentials.GcpCredentials,
            prefect_gcp.credentials.GcpCredentials.load(
                gcp_credentials_block_name,
                _sync=True,  # pyright: ignore[reportCallIssue] _sync is needed to prevent incidental async
            ),
        )
    else:
        logger.debug("Loading default GCP credentials")
        gcp_credentials = prefect_gcp.credentials.GcpCredentials()

    gcs_client = gcp_credentials.get_cloud_storage_client()

    with tempfile.TemporaryDirectory(prefix="prefect-bundle-") as tmp_dir:
        local_path = Path(tmp_dir) / os.path.basename(key)

        try:
            logger.debug(
                "Downloading bundle from GCS bucket %s with key %s to local path %s",
                bucket,
                key,
                local_path,
            )
            gcs_client.bucket(bucket).blob(key).download_to_file(str(local_path))  # pyright: ignore[reportUnknownMemberType] Incomplete type hints
            bundle = from_json(local_path.read_bytes())

            logger.debug("Executing bundle")
            run_coro_as_sync(prefect.runner.Runner().execute_bundle(bundle))
        except Exception as e:
            raise RuntimeError(f"Failed to download bundle from GCS: {e}")


def _cli_wrapper(
    bucket: str = typer.Option(...),
    key: str = typer.Option(...),
    gcp_credentials_block_name: Optional[str] = typer.Option(None),
) -> None:
    execute_bundle_from_gcs(bucket, key, gcp_credentials_block_name)


if __name__ == "__main__":
    typer.run(_cli_wrapper)



import json
from pathlib import Path
from unittest.mock import AsyncMock, MagicMock

import pytest
from google.api_core.exceptions import NotFound as ApiCoreNotFound
from google.cloud.aiplatform_v1.types.job_state import JobState
from google.cloud.exceptions import NotFound
from prefect_gcp.credentials import GcpCredentials

from prefect.settings import PREFECT_LOGGING_TO_API_ENABLED, temporary_settings
from prefect.testing.utilities import prefect_test_harness


@pytest.fixture(scope="session", autouse=True)
def prefect_db():
    with prefect_test_harness():
        yield


@pytest.fixture(scope="session", autouse=True)
def disable_logging():
    with temporary_settings({PREFECT_LOGGING_TO_API_ENABLED: False}):
        yield


@pytest.fixture
def google_auth(monkeypatch):
    google_auth_mock = MagicMock()
    default_credentials_mock = MagicMock(
        client_id="my_client_id", quota_project_id="my_project"
    )
    google_auth_mock.default.side_effect = lambda *args, **kwargs: (
        default_credentials_mock,
        "my_project",
    )
    monkeypatch.setattr("google.auth", google_auth_mock)
    return google_auth_mock


@pytest.fixture
def google_auth_no_quota_project(monkeypatch):
    google_auth_mock = MagicMock()
    default_credentials_mock = MagicMock(
        client_id="my_client_id", quota_project_id=None
    )
    google_auth_mock.default.side_effect = lambda *args, **kwargs: (
        default_credentials_mock,
        "my_project",
    )
    monkeypatch.setattr("google.auth", google_auth_mock)
    return google_auth_mock


@pytest.fixture
def oauth2_credentials(monkeypatch):
    CredentialsMock = MagicMock()
    CredentialsMock.from_service_account_info.side_effect = (
        lambda json, scopes: MagicMock(scopes=scopes, **json)
    )
    CredentialsMock.from_service_account_file.side_effect = lambda file, scopes: file
    monkeypatch.setattr("prefect_gcp.credentials.Credentials", CredentialsMock)


class Blob:
    def __init__(self, name):
        self.name = name

    def download_to_filename(self, filename, **kwargs):
        Path(filename).write_text("abcdef")


class CloudStorageClient(MagicMock):
    def __init__(self, credentials=None, project=None, *args, **kwargs):
        super().__init__()
        self.credentials = credentials
        self.project = project

    def create_bucket(self, bucket, location=None, **create_kwargs):
        bucket_obj = MagicMock(bucket=bucket)
        bucket_obj.name = "my-bucket"
        bucket_obj.location = location
        for key, value in create_kwargs.items():
            setattr(bucket_obj, key, value)
        return bucket_obj

    def get_bucket(self, bucket):
        blob_obj = MagicMock()
        blob_obj.download_as_bytes.return_value = b"bytes"
        blob_obj.download_to_file.side_effect = (
            lambda file_obj, **kwargs: file_obj.write(b"abcdef")
        )
        blob_obj.download_to_filename.side_effect = lambda filename, **kwargs: Path(
            filename
        ).write_text("abcdef")
        bucket_obj = MagicMock(bucket=bucket)
        bucket_obj.name = "my-bucket"
        bucket_obj.blob.side_effect = lambda blob, **kwds: blob_obj
        return bucket_obj

    def list_blobs(self, bucket, prefix=None):
        blob_obj = Blob(name="blob.txt")
        blob_directory = Blob(name="directory/")
        nested_blob_obj = Blob(name="base_folder/nested_blob.txt")
        double_nested_blob_obj = Blob(name="base_folder/sub_folder/nested_blob.txt")
        dotted_folder_blob_obj = Blob(name="dotted.folder/nested_blob.txt")
        blobs = [
            blob_obj,
            blob_directory,
            nested_blob_obj,
            double_nested_blob_obj,
            dotted_folder_blob_obj,
        ]
        if prefix:
            blobs = [blob for blob in blobs if blob.name.startswith(prefix)]
        return blobs


@pytest.fixture
def storage_client(monkeypatch):
    monkeypatch.setattr("prefect_gcp.credentials.StorageClient", CloudStorageClient)


class LoadJob:
    def __init__(self, output, *args, **kwargs):
        self.output = output
        self._client = "client"
        self._completion_lock = "completion_lock"


class BigQueryClient(MagicMock):
    def __init__(self, credentials=None, project=None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.credentials = credentials
        self.project = project

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        pass

    def query(self, query, **kwargs):
        response = MagicMock()
        result = MagicMock()
        result.__iter__.return_value = [query]
        result.to_dataframe.return_value = f"dataframe_{query}"
        result.kwargs = kwargs
        response.result.return_value = result
        response.total_bytes_processed = 10
        return response

    def dataset(self, dataset):
        return MagicMock()

    def table(self, table):
        return MagicMock()

    def get_dataset(self, dataset):
        dataset_obj = MagicMock(table=MagicMock())
        return dataset_obj

    def create_dataset(self, dataset):
        return self.get_dataset(dataset)

    def get_table(self, table):
        raise NotFound("testing")

    def create_table(self, table):
        return table

    def insert_rows_json(self, table, json_rows):
        return json_rows

    def load_table_from_uri(self, uri, *args, **kwargs):
        output = MagicMock()
        output.result.return_value = LoadJob(uri)
        return output

    def load_table_from_file(self, *args, **kwargs):
        output = MagicMock()
        output.result.return_value = LoadJob("file")
        return output


class SecretManagerClient:
    def __init__(self, credentials=None, project=None):
        self.credentials = credentials
        self.project = project
        self._secrets = {}

    def create_secret(self, request=None, parent=None, secret_id=None, **kwds):
        response = MagicMock()
        if request:
            parent = request.parent
            secret_id = request.secret_id
        name = f"{parent}/secrets/{secret_id}"
        response.name = name
        self._secrets[name] = None
        return response

    def add_secret_version(self, request=None, parent=None, payload=None, **kwds):
        response = MagicMock()
        if request:
            parent = request.parent

        if parent not in self._secrets:
            raise ApiCoreNotFound(f"{parent!r} does not exist.")

        response.name = parent
        return response

    def access_secret_version(self, request=None, name=None, **kwds):
        response = MagicMock()
        payload = MagicMock()
        payload.data = "secret_data".encode("utf-8")
        response.payload = payload
        return response

    def delete_secret(self, request=None, name=None, **kwds):
        return name

    def destroy_secret_version(self, name, **kwds):
        return name


@pytest.fixture
def mock_credentials(monkeypatch):
    mock_credentials = MagicMock(name="MockGoogleCredentials")
    mock_authenticated_credentials = MagicMock(token="my-token")
    monkeypatch.setattr(
        "prefect_gcp.credentials.Credentials",  # noqa
        mock_credentials,
    )
    mock_auth = MagicMock()
    mock_auth.default.return_value = (mock_authenticated_credentials, "project")
    monkeypatch.setattr(
        "prefect_gcp.credentials.google.auth",  # noqa
        mock_auth,
    )
    monkeypatch.setattr(
        "prefect_gcp.credentials.Credentials.from_service_account_info.universe_domain",  # noqa
        "fake_universe_domain",
    )
    return mock_credentials


@pytest.fixture
def job_service_client():
    job_service_client_mock = MagicMock()
    custom_run = MagicMock(name="mock_name")
    job_service_client_mock.create_custom_job.return_value = custom_run

    error = MagicMock(message="")
    custom_run_final = MagicMock(
        name="mock_name",
        state=JobState.JOB_STATE_SUCCEEDED,
        error=error,
        display_name="mock_display_name",
    )
    job_service_client_mock.get_custom_job.return_value = custom_run_final
    return job_service_client_mock


@pytest.fixture
def job_service_async_client():
    job_service_client_async_mock = MagicMock()
    custom_run = MagicMock(name="mock_name")
    job_service_client_async_mock.create_custom_job = AsyncMock(return_value=custom_run)

    error = MagicMock(message="")
    custom_run_final = MagicMock(
        name="mock_name",
        state=JobState.JOB_STATE_SUCCEEDED,
        error=error,
        display_name="mock_display_name",
    )
    job_service_client_async_mock.get_custom_job = AsyncMock(
        return_value=custom_run_final
    )
    job_service_client_async_mock.cancel_custom_job = AsyncMock()
    return job_service_client_async_mock


@pytest.fixture()
def service_account_info(monkeypatch):
    monkeypatch.setattr(
        "google.auth.crypt._cryptography_rsa.serialization.load_pem_private_key",
        lambda *args, **kwargs: args[0],
    )
    _service_account_info = {
        "project_id": "my_project",
        "token_uri": "my-token-uri",
        "client_email": "my-client-email",
        "private_key": "my-private-key",
    }
    return _service_account_info


@pytest.fixture()
def service_account_info_json(monkeypatch):
    monkeypatch.setattr(
        "google.auth.crypt._cryptography_rsa.serialization.load_pem_private_key",
        lambda *args, **kwargs: args[0],
    )
    _service_account_info = json.dumps(
        {
            "project_id": "my_project",
            "token_uri": "my-token-uri",
            "client_email": "my-client-email",
            "private_key": "my-private-key",
        }
    )
    return _service_account_info


@pytest.fixture
def gcp_credentials(
    google_auth: MagicMock,
    mock_credentials: MagicMock,
    job_service_client: MagicMock,
    job_service_async_client: MagicMock,
):
    gcp_credentials_mock = MagicMock(spec=GcpCredentials)
    gcp_credentials_mock.service_account_info = None
    gcp_credentials_mock.service_account_info_file = None
    gcp_credentials_mock.project = "gcp_credentials_project"

    gcp_credentials_mock.get_cloud_storage_client.return_value = CloudStorageClient()
    gcp_credentials_mock.get_credentials_from_service_account.return_value = (
        mock_credentials
    )
    gcp_credentials_mock._service_account_email = "my_service_account_email"

    gcp_credentials_mock.job_service_client = job_service_client
    gcp_credentials_mock.job_service_client.__enter__.return_value = job_service_client
    gcp_credentials_mock.job_service_async_client = job_service_async_client
    gcp_credentials_mock.job_service_client.__enter__.return_value = (
        job_service_async_client
    )
    gcp_credentials_mock.get_bigquery_client.return_value = BigQueryClient()
    gcp_credentials_mock.get_secret_manager_client.return_value = SecretManagerClient()
    gcp_credentials_mock.get_job_service_client.return_value = (
        gcp_credentials_mock.job_service_client
    )
    gcp_credentials_mock.get_job_service_async_client.return_value = (
        gcp_credentials_mock.job_service_async_client
    )
    return gcp_credentials_mock



import pytest
from prefect_gcp.secret_manager import (
    GcpSecret,
    create_secret,
    delete_secret,
    delete_secret_version,
    read_secret,
    update_secret,
)

from prefect import flow


def test_create_secret(gcp_credentials):
    @flow
    def test_flow():
        return create_secret("secret_name", gcp_credentials)

    assert test_flow() == "projects/gcp_credentials_project/secrets/secret_name"


@pytest.mark.parametrize("secret_value", ["secret", b"secret_byte"])
def test_update_secret(secret_value, gcp_credentials):
    @flow
    def test_flow():
        create_secret("secret_name", gcp_credentials)
        return update_secret("secret_name", secret_value, gcp_credentials)

    if isinstance(secret_value, str):
        secret_value = secret_value.encode("UTF-8")

    assert test_flow() == "projects/gcp_credentials_project/secrets/secret_name"


def test_read_secret(gcp_credentials):
    @flow
    def test_flow():
        return read_secret("secret_name", gcp_credentials)

    expected = "secret_data"
    assert test_flow() == expected


@pytest.mark.parametrize("project", [None, "override_project"])
def test_delete_secret(project, gcp_credentials):
    secret_name = "secret_name"

    @flow
    def test_flow():
        return delete_secret(secret_name, gcp_credentials, project=project)

    project = project or gcp_credentials.project
    path = f"projects/{project}/secrets/{secret_name}/"
    assert test_flow() == path


@pytest.mark.parametrize("project", [None, "override_project"])
@pytest.mark.parametrize("version_id", ["latest", 1])
def test_delete_secret_version_id(project, version_id, gcp_credentials):
    secret_name = "secret_name"

    @flow
    def test_flow():
        return delete_secret_version(
            secret_name, version_id, gcp_credentials, project=project
        )

    if version_id == "latest":
        with pytest.raises(ValueError):
            test_flow().result()
    else:
        project = project or gcp_credentials.project
        path = f"projects/{project}/secrets/{secret_name}/versions/{version_id}"
        assert test_flow() == path


class TestGcpSecret:
    @pytest.fixture
    def gcp_secret(self, gcp_credentials):
        _gcp_secret = GcpSecret(
            gcp_credentials=gcp_credentials, secret_name="my_secret_name"
        )
        return _gcp_secret

    def test_write_secret(self, gcp_secret):
        expected = "projects/gcp_credentials_project/secrets/my_secret_name"
        actual = gcp_secret.write_secret(secret_data=b"my_secret_data")
        assert actual == expected

    def test_read_secret(self, gcp_secret):
        expected = b"secret_data"
        actual = gcp_secret.read_secret()
        assert actual == expected

    def test_delete_secret(self, gcp_secret):
        expected = "projects/gcp_credentials_project/secrets/my_secret_name"
        actual = gcp_secret.delete_secret()
        assert actual == expected



"""Module handling GCP credentials."""

import functools
import json
from enum import Enum
from pathlib import Path
from threading import Lock
from typing import Any, Dict, Optional, Union

import google.auth
import google.auth.transport.requests
from google.oauth2.service_account import Credentials
from pydantic import Field, field_validator, model_validator

from prefect.blocks.abstract import CredentialsBlock
from prefect.types import SecretDict

try:
    from google.cloud.bigquery import Client as BigQueryClient
except ModuleNotFoundError:
    pass  # will be raised in get_client

try:
    from google.cloud.secretmanager import SecretManagerServiceClient
except ModuleNotFoundError:
    pass

try:
    from google.cloud.storage import Client as StorageClient
except ModuleNotFoundError:
    pass

try:
    from google.cloud.aiplatform.gapic import JobServiceAsyncClient, JobServiceClient
except ModuleNotFoundError:
    pass

try:
    from google.api_core.client_options import ClientOptions, from_dict
except ModuleNotFoundError:
    pass

_LOCK = Lock()


def _raise_help_msg(key: str):
    """
    Raises a helpful error message.

    Args:
        key: the key to access HELP_URLS
    """

    def outer(func):
        """
        Used for decorator.
        """

        @functools.wraps(func)
        def inner(*args, **kwargs):
            """
            Used for decorator.
            """
            try:
                return func(*args, **kwargs)
            except NameError as exc:
                raise ImportError(
                    f"To use prefect_gcp.{key}, install prefect-gcp with the "
                    f"'{key}' extra: `pip install 'prefect_gcp[{key}]'`"
                ) from exc

        return inner

    return outer


class ClientType(Enum):
    CLOUD_STORAGE = "cloud_storage"
    BIGQUERY = "bigquery"
    SECRET_MANAGER = "secret_manager"
    AIPLATFORM = "job_service"  # vertex ai


@functools.lru_cache(maxsize=8, typed=True)
def _get_job_service_async_client_cached(
    ctx, client_options: tuple
) -> "JobServiceAsyncClient":
    """
    Gets an authenticated Job Service async client for Vertex AI.

    Returns:
        An authenticated Job Service async client.
    """
    with _LOCK:
        client_options = dict(client_options)
        credentials = ctx.get_credentials_from_service_account()
        job_service_client = JobServiceAsyncClient(
            credentials=credentials, client_options=client_options
        )
    return job_service_client


class GcpCredentials(CredentialsBlock):
    """
    Block used to manage authentication with GCP. Google authentication is
    handled via the `google.oauth2` module or through the CLI.
    Specify either one of service `account_file` or `service_account_info`; if both
    are not specified, the client will try to detect the credentials following Google's
    [Application Default Credentials](https://cloud.google.com/docs/authentication/application-default-credentials).
    See Google's [Authentication documentation](https://cloud.google.com/docs/authentication#service-accounts)
    for details on inference and recommended authentication patterns.

    Attributes:
        service_account_file: Path to the service account JSON keyfile.
        service_account_info: The contents of the keyfile as a dict.

    Example:
        Load GCP credentials stored in a `GCP Credentials` Block:
        ```python
        from prefect_gcp import GcpCredentials
        gcp_credentials_block = GcpCredentials.load("BLOCK_NAME")
        ```
    """  # noqa

    _logo_url = "https://cdn.sanity.io/images/3ugk85nk/production/10424e311932e31c477ac2b9ef3d53cefbaad708-250x250.png"  # noqa
    _block_type_name = "GCP Credentials"
    _documentation_url = "https://docs.prefect.io/integrations/prefect-gcp"  # noqa: E501

    service_account_file: Optional[Path] = Field(
        default=None, description="Path to the service account JSON keyfile."
    )
    service_account_info: Optional[SecretDict] = Field(
        default=None, description="The contents of the keyfile as a dict."
    )
    project: Optional[str] = Field(
        default=None, description="The GCP project to use for the client."
    )

    _service_account_email: Optional[str] = None

    def __hash__(self):
        return hash(
            (
                hash(self.service_account_file),
                hash(frozenset(self.service_account_info.get_secret_value().items()))
                if self.service_account_info
                else None,
                hash(self.project),
                hash(self._service_account_email),
            )
        )

    @model_validator(mode="after")
    def _provide_one_service_account_source(self):
        """
        Ensure that only a service account file or service account info ias provided.
        """
        if self.service_account_info and self.service_account_file:
            raise ValueError(
                "Only one of service_account_info or service_account_file "
                "can be specified at once"
            )
        return self

    @field_validator("service_account_file")
    @classmethod
    def _check_service_account_file(cls, file):
        """Get full path of provided file and make sure that it exists."""
        if not file:
            return file

        service_account_file = Path(file).expanduser()
        if not service_account_file.exists():
            raise ValueError("The provided path to the service account is invalid")
        return service_account_file

    @field_validator("service_account_info", mode="before")
    @classmethod
    def _convert_json_string_json_service_account_info(cls, value):
        """
        Converts service account info provided as a json formatted string
        to a dictionary
        """
        if isinstance(value, str):
            try:
                service_account_info = json.loads(value)
                return service_account_info
            except Exception:
                raise ValueError("Unable to decode service_account_info")
        else:
            return value

    def block_initialization(self):
        credentials = self.get_credentials_from_service_account()
        if self.project is None:
            if self.service_account_info or self.service_account_file:
                credentials_project = credentials.project_id
            # google.auth.default using gcloud auth application-default login
            elif credentials.quota_project_id:
                credentials_project = credentials.quota_project_id
            # compute-assigned service account via GCP metadata server
            else:
                _, credentials_project = google.auth.default()
            self.project = credentials_project

        if hasattr(credentials, "service_account_email"):
            self._service_account_email = credentials.service_account_email

    def get_credentials_from_service_account(self) -> Credentials:
        """
        Helper method to serialize credentials by using either
        service_account_file or service_account_info.
        """
        if self.service_account_info:
            credentials = Credentials.from_service_account_info(
                self.service_account_info.get_secret_value(),
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        elif self.service_account_file:
            credentials = Credentials.from_service_account_file(
                self.service_account_file,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        else:
            credentials, _ = google.auth.default()
        return credentials

    def get_access_token(self):
        """
        See: https://stackoverflow.com/a/69107745
        Also: https://www.jhanley.com/google-cloud-creating-oauth-access-tokens-for-rest-api-calls/
        """  # noqa
        request = google.auth.transport.requests.Request()
        credentials = self.get_credentials_from_service_account()
        credentials.refresh(request)
        return credentials.token

    def get_client(
        self,
        client_type: Union[str, ClientType],
        **get_client_kwargs: Dict[str, Any],
    ) -> Any:
        """
        Helper method to dynamically get a client type.

        Args:
            client_type: The name of the client to get.
            **get_client_kwargs: Additional keyword arguments to pass to the
                `get_*_client` method.

        Returns:
            An authenticated client.

        Raises:
            ValueError: if the client is not supported.
        """
        if isinstance(client_type, str):
            client_type = ClientType(client_type)
        client_type = client_type.value
        get_client_method = getattr(self, f"get_{client_type}_client")
        return get_client_method(**get_client_kwargs)

    @_raise_help_msg("cloud_storage")
    def get_cloud_storage_client(
        self, project: Optional[str] = None
    ) -> "StorageClient":
        """
        Gets an authenticated Cloud Storage client.

        Args:
            project: Name of the project to use; overrides the base
                class's project if provided.

        Returns:
            An authenticated Cloud Storage client.

        Examples:
            Gets a GCP Cloud Storage client from a path.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_file = "~/.secrets/prefect-service-account.json"
                client = GcpCredentials(
                    service_account_file=service_account_file
                ).get_cloud_storage_client()
            example_get_client_flow()
            ```

            Gets a GCP Cloud Storage client from a dictionary.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_info = {
                    "type": "service_account",
                    "project_id": "project_id",
                    "private_key_id": "private_key_id",
                    "private_key": "private_key",
                    "client_email": "client_email",
                    "client_id": "client_id",
                    "auth_uri": "auth_uri",
                    "token_uri": "token_uri",
                    "auth_provider_x509_cert_url": "auth_provider_x509_cert_url",
                    "client_x509_cert_url": "client_x509_cert_url"
                }
                client = GcpCredentials(
                    service_account_info=service_account_info
                ).get_cloud_storage_client()
            example_get_client_flow()
            ```
        """
        credentials = self.get_credentials_from_service_account()

        # override class project if method project is provided
        project = project or self.project
        storage_client = StorageClient(credentials=credentials, project=project)
        return storage_client

    @_raise_help_msg("bigquery")
    def get_bigquery_client(
        self, project: Optional[str] = None, location: Optional[str] = None
    ) -> "BigQueryClient":
        """
        Gets an authenticated BigQuery client.

        Args:
            project: Name of the project to use; overrides the base
                class's project if provided.
            location: Location to use.

        Returns:
            An authenticated BigQuery client.

        Examples:
            Gets a GCP BigQuery client from a path.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_file = "~/.secrets/prefect-service-account.json"
                client = GcpCredentials(
                    service_account_file=service_account_file
                ).get_bigquery_client()
            example_get_client_flow()
            ```

            Gets a GCP BigQuery client from a dictionary.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_info = {
                    "type": "service_account",
                    "project_id": "project_id",
                    "private_key_id": "private_key_id",
                    "private_key": "private_key",
                    "client_email": "client_email",
                    "client_id": "client_id",
                    "auth_uri": "auth_uri",
                    "token_uri": "token_uri",
                    "auth_provider_x509_cert_url": "auth_provider_x509_cert_url",
                    "client_x509_cert_url": "client_x509_cert_url"
                }
                client = GcpCredentials(
                    service_account_info=service_account_info
                ).get_bigquery_client()

            example_get_client_flow()
            ```
        """
        credentials = self.get_credentials_from_service_account()

        # override class project if method project is provided
        project = project or self.project
        big_query_client = BigQueryClient(
            credentials=credentials, project=project, location=location
        )
        return big_query_client

    @_raise_help_msg("secret_manager")
    def get_secret_manager_client(self) -> "SecretManagerServiceClient":
        """
        Gets an authenticated Secret Manager Service client.

        Returns:
            An authenticated Secret Manager Service client.

        Examples:
            Gets a GCP Secret Manager client from a path.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_file = "~/.secrets/prefect-service-account.json"
                client = GcpCredentials(
                    service_account_file=service_account_file
                ).get_secret_manager_client()
            example_get_client_flow()
            ```

            Gets a GCP Cloud Storage client from a dictionary.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_info = {
                    "type": "service_account",
                    "project_id": "project_id",
                    "private_key_id": "private_key_id",
                    "private_key": "private_key",
                    "client_email": "client_email",
                    "client_id": "client_id",
                    "auth_uri": "auth_uri",
                    "token_uri": "token_uri",
                    "auth_provider_x509_cert_url": "auth_provider_x509_cert_url",
                    "client_x509_cert_url": "client_x509_cert_url"
                }
                client = GcpCredentials(
                    service_account_info=service_account_info
                ).get_secret_manager_client()
            example_get_client_flow()
            ```
        """
        credentials = self.get_credentials_from_service_account()

        # doesn't accept project; must pass in project in tasks
        secret_manager_client = SecretManagerServiceClient(credentials=credentials)
        return secret_manager_client

    @_raise_help_msg("aiplatform")
    def get_job_service_client(
        self, client_options: Union[Dict[str, Any], ClientOptions] = None
    ) -> "JobServiceClient":
        """
        Gets an authenticated Job Service client for Vertex AI.

        Returns:
            An authenticated Job Service client.

        Examples:
            Gets a GCP Job Service client from a path.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_file = "~/.secrets/prefect-service-account.json"
                client = GcpCredentials(
                    service_account_file=service_account_file
                ).get_job_service_client()

            example_get_client_flow()
            ```

            Gets a GCP Cloud Storage client from a dictionary.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_info = {
                    "type": "service_account",
                    "project_id": "project_id",
                    "private_key_id": "private_key_id",
                    "private_key": "private_key",
                    "client_email": "client_email",
                    "client_id": "client_id",
                    "auth_uri": "auth_uri",
                    "token_uri": "token_uri",
                    "auth_provider_x509_cert_url": "auth_provider_x509_cert_url",
                    "client_x509_cert_url": "client_x509_cert_url"
                }
                client = GcpCredentials(
                    service_account_info=service_account_info
                ).get_job_service_client()

            example_get_client_flow()
            ```
        """
        if isinstance(client_options, dict):
            client_options = from_dict(client_options)

        credentials = self.get_credentials_from_service_account()
        return JobServiceClient(credentials=credentials, client_options=client_options)

    @_raise_help_msg("aiplatform")
    def get_job_service_async_client(
        self, client_options: Union[Dict[str, Any], ClientOptions] = None
    ) -> "JobServiceAsyncClient":
        """
        Gets an authenticated Job Service async client for Vertex AI.

        Returns:
            An authenticated Job Service async client.

        Examples:
            Gets a GCP Job Service client from a path.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_file = "~/.secrets/prefect-service-account.json"
                client = GcpCredentials(
                    service_account_file=service_account_file
                ).get_job_service_async_client()

            example_get_client_flow()
            ```

            Gets a GCP Cloud Storage client from a dictionary.
            ```python
            from prefect import flow
            from prefect_gcp.credentials import GcpCredentials

            @flow()
            def example_get_client_flow():
                service_account_info = {
                    "type": "service_account",
                    "project_id": "project_id",
                    "private_key_id": "private_key_id",
                    "private_key": "private_key",
                    "client_email": "client_email",
                    "client_id": "client_id",
                    "auth_uri": "auth_uri",
                    "token_uri": "token_uri",
                    "auth_provider_x509_cert_url": "auth_provider_x509_cert_url",
                    "client_x509_cert_url": "client_x509_cert_url"
                }
                client = GcpCredentials(
                    service_account_info=service_account_info
                ).get_job_service_async_client()

            example_get_client_flow()
            ```
        """
        if isinstance(client_options, dict):
            client_options = from_dict(client_options)

        return _get_job_service_async_client_cached(
            self, tuple(client_options.__dict__.items())
        )



import uuid
from typing import Generator
from unittest.mock import AsyncMock, MagicMock

import pytest
from prefect_gcp.experimental import cloud_run, vertex_ai
from prefect_gcp.workers.cloud_run_v2 import CloudRunWorkerV2
from prefect_gcp.workers.vertex import VertexAIWorker

import prefect
from prefect.client.schemas.actions import WorkPoolCreate
from prefect.client.schemas.objects import WorkPool, WorkPoolStorageConfiguration
from prefect.futures import PrefectFuture


class TestCloudRunDecorator:
    @pytest.fixture
    def mock_submit(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> Generator[AsyncMock, None, None]:
        """Create a mock for the CloudRunWorkerV2.submit method"""
        # Create a mock state
        mock_state = MagicMock(spec=prefect.State)
        mock_state.is_completed.return_value = True
        mock_state.message = "Success"

        # Create a mock future
        mock_future = MagicMock(spec=PrefectFuture)
        mock_future.aresult = AsyncMock(return_value="test_result")
        mock_future.wait_async = AsyncMock()
        mock_future.state = mock_state

        mock = AsyncMock(return_value=mock_future)

        monkeypatch.setattr(CloudRunWorkerV2, "submit", mock)
        yield mock

    @pytest.fixture
    async def work_pool_without_storage_configuration(self):
        async with prefect.get_client() as client:
            work_pool = await client.create_work_pool(
                WorkPoolCreate(
                    name=f"test-{uuid.uuid4()}",
                    base_job_template=CloudRunWorkerV2.get_default_base_job_template(),
                )
            )
            try:
                yield work_pool
            finally:
                await client.delete_work_pool(work_pool.name)

    @pytest.fixture
    async def work_pool_with_storage_configuration(self):
        UPLOAD_STEP = {
            "prefect_mock.experimental.bundles.upload": {
                "requires": "prefect-mock==0.5.5",
                "bucket": "test-bucket",
                "credentials_block_name": "my-creds",
            }
        }

        EXECUTE_STEP = {
            "prefect_mock.experimental.bundles.execute": {
                "requires": "prefect-mock==0.5.5",
                "bucket": "test-bucket",
                "credentials_block_name": "my-creds",
            }
        }

        async with prefect.get_client() as client:
            work_pool = await client.create_work_pool(
                WorkPoolCreate(
                    name=f"test-{uuid.uuid4()}",
                    base_job_template=CloudRunWorkerV2.get_default_base_job_template(),
                    storage_configuration=WorkPoolStorageConfiguration(
                        bundle_execution_step=EXECUTE_STEP,
                        bundle_upload_step=UPLOAD_STEP,
                    ),
                )
            )
            try:
                yield work_pool
            finally:
                await client.delete_work_pool(work_pool.name)

    def test_cloud_run_decorator_sync_flow(self, mock_submit: AsyncMock) -> None:
        """Test that a synchronous flow is correctly decorated and executed"""

        @cloud_run(work_pool="test-pool", memory="2Gi")
        @prefect.flow
        def sync_test_flow(param1: str, param2: str = "default") -> str:
            return f"{param1}-{param2}"

        result = sync_test_flow("test")

        mock_submit.assert_called_once()
        _args, kwargs = mock_submit.call_args
        assert kwargs["parameters"] == {"param1": "test", "param2": "default"}
        assert kwargs["job_variables"] == {"memory": "2Gi"}
        assert result == "test_result"

    async def test_cloud_run_decorator_async_flow(self, mock_submit: AsyncMock) -> None:
        """Test that an asynchronous flow is correctly decorated and executed"""

        @cloud_run(work_pool="test-pool", cpu="1")
        @prefect.flow
        async def async_test_flow(param1: str) -> str:
            return f"async-{param1}"

        result = await async_test_flow("test")

        mock_submit.assert_called_once()
        _args, kwargs = mock_submit.call_args
        assert kwargs["parameters"] == {"param1": "test"}
        assert kwargs["job_variables"] == {"cpu": "1"}
        assert result == "test_result"

    @pytest.mark.usefixtures("mock_submit")
    def test_cloud_run_decorator_return_state(self) -> None:
        """Test that return_state=True returns the state instead of the result"""

        @cloud_run(work_pool="test-pool")
        @prefect.flow
        def test_flow():
            return "completed"

        state = test_flow(return_state=True)

        assert state.is_completed()
        assert state.message == "Success"

    @pytest.mark.usefixtures("mock_submit")
    def test_cloud_run_decorator_preserves_flow_attributes(self) -> None:
        """Test that the decorator preserves the original flow's attributes"""

        @prefect.flow(name="custom-flow-name", description="Custom description")
        def original_flow():
            return "test"

        original_name = original_flow.name
        original_description = original_flow.description

        decorated_flow = cloud_run(work_pool="test-pool")(original_flow)

        assert decorated_flow.name == original_name
        assert decorated_flow.description == original_description

        result = decorated_flow()
        assert result == "test_result"

    def test_submit_method_receives_work_pool_name(
        self, mock_submit: AsyncMock
    ) -> None:
        """Test that the correct work pool name is passed to submit"""

        @cloud_run(work_pool="specific-pool")
        @prefect.flow
        def test_flow():
            return "test"

        test_flow()

        mock_submit.assert_called_once()
        kwargs = mock_submit.call_args.kwargs
        assert "flow" in kwargs
        assert "parameters" in kwargs
        assert "job_variables" in kwargs

    def test_runtime_error_when_work_pool_is_missing_storage_configuration(
        self,
        work_pool_without_storage_configuration: WorkPool,
    ) -> None:
        """Test that a runtime error is raised when the work pool is missing a storage configuration"""

        @cloud_run(work_pool=work_pool_without_storage_configuration.name)
        @prefect.flow
        def test_flow():
            return "test"

        with pytest.raises(
            RuntimeError,
            match=f"Storage is not configured for work pool {work_pool_without_storage_configuration.name!r}. "
            "Please configure storage for the work pool by running `prefect work-pool storage configure`.",
        ):
            test_flow()


class TestVertexAIDecorator:
    @pytest.fixture
    def mock_submit(
        self, monkeypatch: pytest.MonkeyPatch
    ) -> Generator[AsyncMock, None, None]:
        """Create a mock for the VertexAIWorker.submit method"""
        # Create a mock state
        mock_state = MagicMock(spec=prefect.State)
        mock_state.is_completed.return_value = True
        mock_state.message = "Success"

        # Create a mock future
        mock_future = MagicMock(spec=PrefectFuture)
        mock_future.aresult = AsyncMock(return_value="test_result")
        mock_future.wait_async = AsyncMock()
        mock_future.state = mock_state

        mock = AsyncMock(return_value=mock_future)

        monkeypatch.setattr(VertexAIWorker, "submit", mock)
        yield mock

    @pytest.fixture
    async def work_pool_without_storage_configuration(self):
        async with prefect.get_client() as client:
            work_pool = await client.create_work_pool(
                WorkPoolCreate(
                    name=f"test-{uuid.uuid4()}",
                    base_job_template=VertexAIWorker.get_default_base_job_template(),
                )
            )
            try:
                yield work_pool
            finally:
                await client.delete_work_pool(work_pool.name)

    @pytest.fixture
    async def work_pool_with_storage_configuration(self):
        UPLOAD_STEP = {
            "prefect_mock.experimental.bundles.upload": {
                "requires": "prefect-mock==0.5.5",
                "bucket": "test-bucket",
                "credentials_block_name": "my-creds",
            }
        }

        EXECUTE_STEP = {
            "prefect_mock.experimental.bundles.execute": {
                "requires": "prefect-mock==0.5.5",
                "bucket": "test-bucket",
                "credentials_block_name": "my-creds",
            }
        }

        async with prefect.get_client() as client:
            work_pool = await client.create_work_pool(
                WorkPoolCreate(
                    name=f"test-{uuid.uuid4()}",
                    base_job_template=VertexAIWorker.get_default_base_job_template(),
                    storage_configuration=WorkPoolStorageConfiguration(
                        bundle_execution_step=EXECUTE_STEP,
                        bundle_upload_step=UPLOAD_STEP,
                    ),
                )
            )
            try:
                yield work_pool
            finally:
                await client.delete_work_pool(work_pool.name)

    def test_cloud_run_decorator_sync_flow(self, mock_submit: AsyncMock) -> None:
        """Test that a synchronous flow is correctly decorated and executed"""

        @vertex_ai(work_pool="test-pool", memory="2Gi")
        @prefect.flow
        def sync_test_flow(param1: str, param2: str = "default") -> str:
            return f"{param1}-{param2}"

        result = sync_test_flow("test")

        mock_submit.assert_called_once()
        _args, kwargs = mock_submit.call_args
        assert kwargs["parameters"] == {"param1": "test", "param2": "default"}
        assert kwargs["job_variables"] == {"memory": "2Gi"}
        assert result == "test_result"

    async def test_cloud_run_decorator_async_flow(self, mock_submit: AsyncMock) -> None:
        """Test that an asynchronous flow is correctly decorated and executed"""

        @vertex_ai(work_pool="test-pool", cpu="1")
        @prefect.flow
        async def async_test_flow(param1: str) -> str:
            return f"async-{param1}"

        result = await async_test_flow("test")

        mock_submit.assert_called_once()
        _args, kwargs = mock_submit.call_args
        assert kwargs["parameters"] == {"param1": "test"}
        assert kwargs["job_variables"] == {"cpu": "1"}
        assert result == "test_result"

    @pytest.mark.usefixtures("mock_submit")
    def test_cloud_run_decorator_return_state(self) -> None:
        """Test that return_state=True returns the state instead of the result"""

        @vertex_ai(work_pool="test-pool")
        @prefect.flow
        def test_flow():
            return "completed"

        state = test_flow(return_state=True)

        assert state.is_completed()
        assert state.message == "Success"

    @pytest.mark.usefixtures("mock_submit")
    def test_vertex_ai_decorator_preserves_flow_attributes(self) -> None:
        """Test that the decorator preserves the original flow's attributes"""

        @prefect.flow(name="custom-flow-name", description="Custom description")
        def original_flow():
            return "test"

        original_name = original_flow.name
        original_description = original_flow.description

        decorated_flow = vertex_ai(work_pool="test-pool")(original_flow)

        assert decorated_flow.name == original_name
        assert decorated_flow.description == original_description

        result = decorated_flow()
        assert result == "test_result"

    def test_submit_method_receives_work_pool_name(
        self, mock_submit: AsyncMock
    ) -> None:
        """Test that the correct work pool name is passed to submit"""

        @vertex_ai(work_pool="specific-pool")
        @prefect.flow
        def test_flow():
            return "test"

        test_flow()

        mock_submit.assert_called_once()
        kwargs = mock_submit.call_args.kwargs
        assert "flow" in kwargs
        assert "parameters" in kwargs
        assert "job_variables" in kwargs

    def test_runtime_error_when_work_pool_is_missing_storage_configuration(
        self,
        work_pool_without_storage_configuration: WorkPool,
    ) -> None:
        """Test that a runtime error is raised when the work pool is missing a storage configuration"""

        @vertex_ai(work_pool=work_pool_without_storage_configuration.name)
        @prefect.flow
        def test_flow():
            return "test"

        with pytest.raises(
            RuntimeError,
            match=f"Storage is not configured for work pool {work_pool_without_storage_configuration.name!r}. "
            "Please configure storage for the work pool by running `prefect work-pool storage configure`.",
        ):
            test_flow()



from __future__ import annotations

import uuid
from unittest.mock import Mock

import pydantic
import pytest
from prefect_gcp.credentials import GcpCredentials
from prefect_gcp.models.cloud_run_v2 import SecretKeySelector
from prefect_gcp.utilities import slugify_name
from prefect_gcp.workers.cloud_run import (
    CloudRunWorker,
    CloudRunWorkerJobConfiguration,
    CloudRunWorkerResult,
)

from prefect.client.schemas.objects import FlowRun
from prefect.server.schemas.actions import DeploymentCreate
from prefect.utilities.dockerutils import get_prefect_image_name
from prefect.utilities.schema_tools.validation import (
    ValidationError as PrefectValidationError,
)


@pytest.fixture(autouse=True)
def mock_credentials(gcp_credentials):
    yield


@pytest.fixture
def jobs_body():
    return {
        "apiVersion": "run.googleapis.com/v1",
        "kind": "Job",
        "metadata": {
            "annotations": {
                # See: https://cloud.google.com/run/docs/troubleshooting#launch-stage-validation  # noqa
                "run.googleapis.com/launch-stage": "BETA",
            },
        },
        "spec": {  # JobSpec
            "template": {  # ExecutionTemplateSpec
                "spec": {  # ExecutionSpec
                    "template": {  # TaskTemplateSpec
                        "spec": {"containers": [{}]},  # TaskSpec
                    },
                },
            },
        },
    }


@pytest.fixture
def flow_run():
    return FlowRun(flow_id=uuid.uuid4(), name="my-flow-run-name")


@pytest.fixture
def cloud_run_worker_job_config(service_account_info, jobs_body):
    return CloudRunWorkerJobConfiguration(
        name="my-job-name",
        region="middle-earth2",
        job_body=jobs_body,
        credentials=GcpCredentials(service_account_info=service_account_info),
    )


@pytest.fixture
def cloud_run_worker_job_config_noncompliant_name(service_account_info, jobs_body):
    return CloudRunWorkerJobConfiguration(
        name="MY_JOB_NAME",
        region="middle-earth2",
        job_body=jobs_body,
        credentials=GcpCredentials(service_account_info=service_account_info),
    )


class TestCloudRunWorkerJobConfiguration:
    def test_job_name(self, cloud_run_worker_job_config):
        cloud_run_worker_job_config.job_body["metadata"]["name"] = "my-job-name"
        assert cloud_run_worker_job_config.job_name == "my-job-name"

    def test_job_name_is_slug(self, cloud_run_worker_job_config_noncompliant_name):
        cloud_run_worker_job_config_noncompliant_name._populate_name_if_not_present()
        assert cloud_run_worker_job_config_noncompliant_name.job_name[
            :-33
        ] == slugify_name("MY_JOB_NAME")

    def test_populate_envs(
        self,
        cloud_run_worker_job_config,
    ):
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]

        assert "env" not in container
        assert cloud_run_worker_job_config.env == {}
        cloud_run_worker_job_config.env = {"a": "b"}

        cloud_run_worker_job_config._populate_envs()

        assert "env" in container
        assert len(container["env"]) == 1
        assert container["env"][0]["name"] == "a"
        assert container["env"][0]["value"] == "b"

    def test_populate_envs_with_prefect_api_key_secret(
        self,
        cloud_run_worker_job_config,
    ):
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]

        # Set up environment with API key
        cloud_run_worker_job_config.env = {
            "PREFECT_API_KEY": "test-api-key",
            "OTHER_VAR": "other-value",
        }
        cloud_run_worker_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="my-secret", version="latest"
        )

        cloud_run_worker_job_config._populate_envs()

        assert "env" in container
        assert len(container["env"]) == 2

        # Should have OTHER_VAR as regular env var
        regular_env = [env for env in container["env"] if env["name"] == "OTHER_VAR"][0]
        assert regular_env["value"] == "other-value"

        # Should have PREFECT_API_KEY as secret
        secret_env = [
            env for env in container["env"] if env["name"] == "PREFECT_API_KEY"
        ][0]
        assert "value" not in secret_env  # Should not have plaintext value
        assert "valueFrom" in secret_env
        assert secret_env["valueFrom"]["secretKeyRef"]["name"] == "my-secret"
        assert secret_env["valueFrom"]["secretKeyRef"]["key"] == "latest"

    def test_populate_envs_with_prefect_api_auth_string_secret(
        self,
        cloud_run_worker_job_config,
    ):
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]

        # Set up environment with auth string
        cloud_run_worker_job_config.env = {
            "PREFECT_API_AUTH_STRING": "test-auth-string",
            "OTHER_VAR": "other-value",
        }
        cloud_run_worker_job_config.prefect_api_auth_string_secret = SecretKeySelector(
            secret="my-auth-secret", version="latest"
        )

        cloud_run_worker_job_config._populate_envs()

        assert "env" in container
        assert len(container["env"]) == 2

        # Should have OTHER_VAR as regular env var
        regular_env = [env for env in container["env"] if env["name"] == "OTHER_VAR"][0]
        assert regular_env["value"] == "other-value"

        # Should have PREFECT_API_AUTH_STRING as secret
        secret_env = [
            env for env in container["env"] if env["name"] == "PREFECT_API_AUTH_STRING"
        ][0]
        assert "value" not in secret_env  # Should not have plaintext value
        assert "valueFrom" in secret_env
        assert secret_env["valueFrom"]["secretKeyRef"]["name"] == "my-auth-secret"
        assert secret_env["valueFrom"]["secretKeyRef"]["key"] == "latest"

    def test_populate_envs_with_both_secrets(
        self,
        cloud_run_worker_job_config,
    ):
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]

        # Set up environment with both API key and auth string
        cloud_run_worker_job_config.env = {
            "PREFECT_API_KEY": "test-api-key",
            "PREFECT_API_AUTH_STRING": "test-auth-string",
            "OTHER_VAR": "other-value",
        }
        cloud_run_worker_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="my-secret", version="latest"
        )
        cloud_run_worker_job_config.prefect_api_auth_string_secret = SecretKeySelector(
            secret="my-auth-secret", version="latest"
        )

        cloud_run_worker_job_config._populate_envs()

        assert "env" in container
        assert len(container["env"]) == 3

        # Should have OTHER_VAR as regular env var
        regular_env = [env for env in container["env"] if env["name"] == "OTHER_VAR"][0]
        assert regular_env["value"] == "other-value"

        # Should have PREFECT_API_KEY as secret
        api_key_env = [
            env for env in container["env"] if env["name"] == "PREFECT_API_KEY"
        ][0]
        assert "value" not in api_key_env
        assert api_key_env["valueFrom"]["secretKeyRef"]["name"] == "my-secret"
        assert api_key_env["valueFrom"]["secretKeyRef"]["key"] == "latest"

        # Should have PREFECT_API_AUTH_STRING as secret
        auth_string_env = [
            env for env in container["env"] if env["name"] == "PREFECT_API_AUTH_STRING"
        ][0]
        assert "value" not in auth_string_env
        assert auth_string_env["valueFrom"]["secretKeyRef"]["name"] == "my-auth-secret"
        assert auth_string_env["valueFrom"]["secretKeyRef"]["key"] == "latest"

    def test_populate_envs_with_secret_but_no_env_var(
        self,
        cloud_run_worker_job_config,
    ):
        """Test that secret is added even when the corresponding env var doesn't exist"""
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]

        # No PREFECT_API_KEY in env, but secret is configured
        cloud_run_worker_job_config.env = {"OTHER_VAR": "other-value"}
        cloud_run_worker_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="my-secret", version="latest"
        )

        cloud_run_worker_job_config._populate_envs()

        assert "env" in container
        assert len(container["env"]) == 2

        # Should have PREFECT_API_KEY from secret
        secret_env = [
            env for env in container["env"] if env["name"] == "PREFECT_API_KEY"
        ][0]
        assert secret_env["valueFrom"]["secretKeyRef"]["name"] == "my-secret"
        assert secret_env["valueFrom"]["secretKeyRef"]["key"] == "latest"

    def test_populate_envs_no_warning_when_secrets_configured(
        self, cloud_run_worker_job_config, caplog
    ):
        """Test that no warnings are logged when secrets are properly configured"""
        # Set up environment with both plaintext and secret
        cloud_run_worker_job_config.env = {
            "PREFECT_API_KEY": "test-api-key",
            "PREFECT_API_AUTH_STRING": "test-auth-string",
            "OTHER_VAR": "other-value",
        }
        cloud_run_worker_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="my-secret", version="latest"
        )
        cloud_run_worker_job_config.prefect_api_auth_string_secret = SecretKeySelector(
            secret="my-auth-secret", version="latest"
        )

        with caplog.at_level("WARNING"):
            cloud_run_worker_job_config._populate_envs()

        # Check that no warnings were logged since secrets are configured
        assert len(caplog.records) == 0

    def test_populate_envs_logs_warning_when_plaintext_without_secret(
        self, cloud_run_worker_job_config, caplog, flow_run
    ):
        """Test that warnings are logged when plaintext credentials are provided without secrets"""
        # Set up environment with plaintext credentials but no secrets
        cloud_run_worker_job_config.env = {
            "PREFECT_API_KEY": "test-api-key",
            "PREFECT_API_AUTH_STRING": "test-auth-string",
            "OTHER_VAR": "other-value",
        }
        # No secrets configured
        cloud_run_worker_job_config.prefect_api_key_secret = None
        cloud_run_worker_job_config.prefect_api_auth_string_secret = None

        cloud_run_worker_job_config.prepare_for_flow_run(
            flow_run=flow_run,
        )

        assert (
            "PREFECT_API_KEY is provided as a plaintext environment variable"
            in caplog.text
        )
        assert (
            "PREFECT_API_AUTH_STRING is provided as a plaintext environment variable"
            in caplog.text
        )
        assert (
            "consider providing it as a secret using 'prefect_api_key_secret' in your base job template"
            in caplog.text
        )
        assert (
            "consider providing it as a secret using 'prefect_api_auth_string_secret' in your base job template"
            in caplog.text
        )

    def test_populate_image_if_not_present(self, cloud_run_worker_job_config):
        cloud_run_worker_job_config._populate_image_if_not_present()

        assert (
            cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
                "template"
            ]["spec"]["containers"][0]["image"]
            == f"docker.io/{get_prefect_image_name()}"
        )

    def test_populate_image_doesnt_overwrite(self, cloud_run_worker_job_config):
        image = "my-first-image"
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]
        container["image"] = image

        cloud_run_worker_job_config._populate_image_if_not_present()

        assert container["image"] == image

    def test_populate_name_if_not_present(self, cloud_run_worker_job_config):
        metadata = cloud_run_worker_job_config.job_body["metadata"]

        assert "name" not in metadata

        cloud_run_worker_job_config._populate_name_if_not_present()

        assert "name" in metadata
        assert metadata["name"][:-33] == cloud_run_worker_job_config.name

    def test_job_name_different_after_retry(self, cloud_run_worker_job_config):
        metadata = cloud_run_worker_job_config.job_body["metadata"]

        assert "name" not in metadata

        cloud_run_worker_job_config._populate_name_if_not_present()
        job_name_1 = metadata.pop("name")

        cloud_run_worker_job_config._populate_name_if_not_present()
        job_name_2 = metadata.pop("name")

        assert job_name_1[:-33] == job_name_2[:-33]
        assert job_name_1 != job_name_2

    def test_populate_or_format_command_doesnt_exist(self, cloud_run_worker_job_config):
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]

        assert "command" not in container

        cloud_run_worker_job_config._populate_or_format_command()

        assert "command" in container
        assert container["command"] == ["prefect", "flow-run", "execute"]

    def test_populate_or_format_command_already_exists(
        self, cloud_run_worker_job_config
    ):
        command = "my command and args"
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]
        container["command"] = command

        cloud_run_worker_job_config._populate_or_format_command()

        assert "command" in container
        assert container["command"] == command.split()

    def test_format_args_if_present(self, cloud_run_worker_job_config):
        args = "my args"
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]
        container["args"] = args

        cloud_run_worker_job_config._format_args_if_present()

        assert "args" in container
        assert container["args"] == args.split()

    def test_format_args_if_present_no_args(self, cloud_run_worker_job_config):
        container = cloud_run_worker_job_config.job_body["spec"]["template"]["spec"][
            "template"
        ]["spec"]["containers"][0]
        assert "args" not in container

        cloud_run_worker_job_config._format_args_if_present()

        assert "args" not in container

    def test_populate_name_doesnt_overwrite(
        self, cloud_run_worker_job_config, flow_run
    ):
        name = "my-name"
        metadata = cloud_run_worker_job_config.job_body["metadata"]
        metadata["name"] = name

        cloud_run_worker_job_config._populate_name_if_not_present()

        assert "name" in metadata
        assert metadata["name"] != flow_run.name
        assert metadata["name"] == name

    async def test_validates_against_an_empty_job_body(self):
        template = CloudRunWorker.get_default_base_job_template()
        template["job_configuration"]["job_body"] = {}
        template["job_configuration"]["region"] = "test-region1"

        with pytest.raises(pydantic.ValidationError) as excinfo:
            await CloudRunWorkerJobConfiguration.from_template_and_values(template, {})

        assert len(err := excinfo.value.errors()) == 1
        assert err[0]["loc"] == ("job_body",)
        assert "Value error, Job is missing required attributes" in err[0]["msg"]
        assert all(
            key in err[0]["msg"]
            for key in ["/apiVersion", "/kind", "/metadata", "/spec"]
        )

    async def test_validates_for_a_job_body_missing_deeper_attributes(self):
        template = CloudRunWorker.get_default_base_job_template()
        template["job_configuration"]["job_body"] = {
            "apiVersion": "run.googleapis.com/v1",
            "kind": "Job",
            "metadata": {},
            "spec": {"template": {"spec": {"template": {"spec": {}}}}},
        }
        template["job_configuration"]["region"] = "test-region1"

        with pytest.raises(pydantic.ValidationError) as excinfo:
            await CloudRunWorkerJobConfiguration.from_template_and_values(template, {})

        assert len(err := excinfo.value.errors()) == 1
        assert err[0]["loc"] == ("job_body",)
        assert "Job is missing required attributes" in err[0]["msg"]
        assert "/metadata/annotations" in err[0]["msg"]
        assert "/spec/template/spec/template/spec/containers" in err[0]["msg"]

    async def test_validates_for_a_job_with_incompatible_values(self):
        """We should give a human-friendly error when the user provides a custom Job
        manifest that is attempting to change required values."""
        template = CloudRunWorker.get_default_base_job_template()
        template["job_configuration"]["region"] = "test-region1"
        template["job_configuration"]["job_body"] = {
            "apiVersion": "v1",
            "kind": "NOTAJOB",
            "metadata": {
                "annotations": {
                    "run.googleapis.com/launch-stage": "NOTABETA",
                },
            },
            "spec": {  # JobSpec
                "template": {  # ExecutionTemplateSpec
                    "spec": {  # ExecutionSpec
                        "template": {  # TaskTemplateSpec
                            "spec": {"containers": [{}]},  # TaskSpec
                        },
                    },
                },
            },
        }

        with pytest.raises(pydantic.ValidationError) as excinfo:
            await CloudRunWorkerJobConfiguration.from_template_and_values(template, {})

        assert len(err := excinfo.value.errors()) == 1
        assert err[0]["loc"] == ("job_body",)
        assert "/apiVersion must have value 'run.googleapis.com/v1'" in err[0]["msg"]
        assert "/kind must have value 'Job'" in err[0]["msg"]
        assert (
            "/metadata/annotations/run.googleapis.com~1launch-stage must have value 'BETA'"
            in err[0]["msg"]
        )


class TestCloudRunWorkerValidConfiguration:
    @pytest.mark.parametrize("cpu", ["1", "100", "100m", "1500m"])
    def test_invalid_cpu_string(self, cpu):
        deployment = DeploymentCreate(
            name="my-deployment",
            flow_id=uuid.uuid4(),
            job_variables={"region": "test-region1", "cpu": cpu},
        )
        with pytest.raises(PrefectValidationError) as exc:
            deployment.check_valid_configuration(
                CloudRunWorker.get_default_base_job_template()
            )

        assert f"'{cpu}' does not match '^(\\\\d*000)m$'" in str(exc.value)

    @pytest.mark.parametrize("cpu", ["1000m", "2000m", "3000m"])
    def test_valid_cpu_string(self, cpu):
        deployment = DeploymentCreate(
            name="my-deployment",
            flow_id=uuid.uuid4(),
            job_variables={"region": "test-region1", "cpu": cpu},
        )
        deployment.check_valid_configuration(
            CloudRunWorker.get_default_base_job_template()
        )

    @pytest.mark.parametrize("memory", ["1", "100", "100Gigs"])
    def test_invalid_memory_string(self, memory):
        deployment = DeploymentCreate(
            name="my-deployment",
            flow_id=uuid.uuid4(),
            job_variables={"region": "test-region1", "memory": memory},
        )
        with pytest.raises(PrefectValidationError) as exc:
            deployment.check_valid_configuration(
                CloudRunWorker.get_default_base_job_template()
            )
        assert f"'{memory}' does not match '^\\\\d+(?:G|Gi|M|Mi)$'" in str(exc.value)

    @pytest.mark.parametrize("memory", ["512G", "512Gi", "512M", "512Mi"])
    def test_valid_memory_string(self, memory):
        deployment = DeploymentCreate(
            name="my-deployment",
            flow_id=uuid.uuid4(),
            job_variables={"region": "test-region1", "memory": memory},
        )
        deployment.check_valid_configuration(
            CloudRunWorker.get_default_base_job_template()
        )


executions_return_value = {
    "metadata": {"name": "test-name", "namespace": "test-namespace"},
    "spec": {"MySpec": "spec"},
    "status": {"logUri": "test-log-uri"},
}

jobs_return_value = {
    "metadata": {"name": "Test", "namespace": "test-namespace"},
    "spec": {"MySpec": "spec"},
    "status": {
        "conditions": [{"type": "Ready", "dog": "cat"}],
        "latestCreatedExecution": {"puppy": "kitty"},
    },
}


@pytest.fixture
def mock_client(monkeypatch, mock_credentials):
    m = Mock(name="MockClient")

    def mock_enter(m, *args, **kwargs):
        return m

    def mock_exit(m, *args, **kwargs):
        pass

    m.__enter__ = mock_enter
    m.__exit__ = mock_exit

    def get_mock_client(*args, **kwargs):
        return m

    monkeypatch.setattr(
        "prefect_gcp.workers.cloud_run.CloudRunWorker._get_client",
        get_mock_client,
    )

    return m


class MockExecution(Mock):
    call_count = 0

    def __init__(self, succeeded=False, *args, **kwargs):
        super().__init__()
        self.log_uri = "test_uri"
        self._succeeded = succeeded

    def is_running(self):
        MockExecution.call_count += 1

        if self.call_count > 2:
            return False
        return True

    def condition_after_completion(self):
        return {"message": "test"}

    def succeeded(self):
        return self._succeeded

    @classmethod
    def get(cls, *args, **kwargs):
        return cls()


def list_mock_calls(mock_client, assigned_calls=0):
    calls = []
    for call in mock_client.mock_calls:
        # mock `call.jobs().get()` results in two calls: `call.jobs()` and
        # `call.jobs().get()`, so we want to remove the first, smaller
        # call.
        if len(str(call).split(".")) > 2:
            calls.append(str(call))
    # assigning a return value to a call results in initial
    # mock calls which are not actually made
    actual_calls = calls[assigned_calls:]

    return actual_calls


class TestCloudRunWorker:
    job_ready = {
        "metadata": {"name": "Test", "namespace": "test-namespace"},
        "spec": {"MySpec": "spec"},
        "status": {
            "conditions": [{"type": "Ready", "status": "True"}],
            "latestCreatedExecution": {"puppy": "kitty"},
        },
    }
    execution_ready = {
        "metadata": {"name": "test-name", "namespace": "test-namespace"},
        "spec": {"MySpec": "spec"},
        "status": {"logUri": "test-log-uri"},
    }
    execution_complete_and_succeeded = {
        "metadata": {"name": "test-name", "namespace": "test-namespace"},
        "spec": {"MySpec": "spec"},
        "status": {
            "logUri": "test-log-uri",
            "completionTime": "Done!",
            "conditions": [{"type": "Completed", "status": "True"}],
        },
    }
    execution_not_found = {
        "metadata": {"name": "test-name", "namespace": "test-namespace"},
        "spec": {"MySpec": "spec"},
        "status": {
            "logUri": "test-log-uri",
            "completionTime": "Done!",
            "conditions": [
                {"type": "Completed", "status": "False", "message": "Not found"}
            ],
        },
    }
    execution_complete_and_failed = {
        "metadata": {"name": "test-name", "namespace": "test-namespace"},
        "spec": {"MySpec": "spec"},
        "status": {
            "logUri": "test-log-uri",
            "completionTime": "Done!",
            "conditions": [
                {"type": "Completed", "status": "False", "message": "test failure"}
            ],
        },
    }

    @pytest.mark.parametrize(
        "keep_job,expected_calls",
        [
            (
                True,
                [
                    "call.jobs().create",
                    "call.jobs().create().execute()",
                ],
            ),
            (
                False,
                [
                    "call.jobs().create",
                    "call.jobs().create().execute()",
                    "call.jobs().get",
                    "call.jobs().get().execute()",
                    "call.jobs().run",
                    "call.jobs().run().execute()",
                ],
            ),
        ],
    )
    async def test_happy_path_api_calls_made_correctly(
        self,
        mock_client,
        cloud_run_worker_job_config,
        keep_job,
        flow_run,
        expected_calls,
    ):
        """Expected behavior:
        Happy path:
        - A call to Job.create and execute
        - A call to Job.get and execute (check status)
        - A call to Job.run and execute (start the job when status is ready)
        - A call to Executions.get and execute (see the status of the Execution)
        - A call to Job.delete and execute if `keep_job` False, otherwise no call
        """
        async with CloudRunWorker("my-work-pool") as cloud_run_worker:
            cloud_run_worker_job_config.keep_job = keep_job
            cloud_run_worker_job_config.prepare_for_flow_run(flow_run, None, None)
            mock_client.jobs().get().execute.return_value = self.job_ready
            mock_client.jobs().run().execute.return_value = self.job_ready
            mock_client.executions().get().execute.return_value = (
                self.execution_complete_and_succeeded
            )
            await cloud_run_worker.run(flow_run, cloud_run_worker_job_config)
            calls = list_mock_calls(mock_client, 3)

            for call, expected_call in zip(calls, expected_calls):
                assert call.startswith(expected_call)

    async def test_happy_path_result(
        self, mock_client, cloud_run_worker_job_config, flow_run
    ):
        """Expected behavior: returns a CloudrunJobResult with status_code 0"""
        async with CloudRunWorker("my-work-pool") as cloud_run_worker:
            cloud_run_worker_job_config.keep_job = True
            cloud_run_worker_job_config.prepare_for_flow_run(flow_run, None, None)
            mock_client.jobs().get().execute.return_value = self.job_ready
            mock_client.jobs().run().execute.return_value = self.job_ready
            mock_client.executions().get().execute.return_value = (
                self.execution_complete_and_succeeded
            )
            res = await cloud_run_worker.run(flow_run, cloud_run_worker_job_config)
            assert isinstance(res, CloudRunWorkerResult)
            assert res.status_code == 0

    @pytest.mark.parametrize(
        "keep_job,expected_calls",
        [
            (
                True,
                [
                    "call.jobs().create",
                    "call.jobs().create().execute()",
                ],
            ),
            (
                False,
                [
                    "call.jobs().create",
                    "call.jobs().create().execute()",
                    "call.jobs().delete",
                    "call.jobs().delete().execute()",
                ],
            ),
        ],
    )
    async def test_behavior_called_when_job_get_fails(
        self,
        monkeypatch,
        mock_client,
        cloud_run_worker_job_config,
        flow_run,
        keep_job,
        expected_calls,
    ):
        """Expected behavior:
        When job create is called, but there is a subsequent exception on a get,
        there should be a delete call for that job if `keep_job` is False
        """
        async with CloudRunWorker("my-work-pool") as cloud_run_worker:
            cloud_run_worker_job_config.keep_job = keep_job

            # Getting job will raise error
            def raise_exception(*args, **kwargs):
                raise Exception("This is an intentional exception")

            monkeypatch.setattr("prefect_gcp.utilities.Job.get", raise_exception)

            with pytest.raises(Exception):
                await cloud_run_worker.run(flow_run, cloud_run_worker_job_config)
            calls = list_mock_calls(mock_client, 0)

            for call, expected_call in zip(calls, expected_calls):
                assert call.startswith(expected_call)

    # Test that RuntimeError raised if something happens with execution
    @pytest.mark.parametrize(
        "keep_job,expected_calls",
        [
            (
                True,
                [
                    "call.jobs().create",
                    "call.jobs().create().execute()",
                    "call.jobs().get",
                    "call.jobs().get().execute()",
                    "call.jobs().run",
                    "call.jobs().run().execute()",
                ],
            ),
            (
                False,
                [
                    "call.jobs().create",
                    "call.jobs().create().execute()",
                    "call.jobs().get",
                    "call.jobs().get().execute()",
                    "call.jobs().run",
                    "call.jobs().run().execute()",
                    "call.jobs().delete",
                    "call.jobs().delete().execute()",
                ],
            ),
        ],
    )
    async def test_behavior_called_when_execution_get_fails(
        self,
        monkeypatch,
        mock_client,
        cloud_run_worker_job_config,
        flow_run,
        keep_job,
        expected_calls,
    ):
        """Expected behavior:
        Job creation is called successfully, the job is found when `get` is called,
        job `run` is called, but the execution fails, there should be a delete
        call for that job if `keep_job` is False, and an exception should be raised.
        """
        async with CloudRunWorker("my-work-pool") as cloud_run_worker:
            cloud_run_worker_job_config.keep_job = keep_job
            mock_client.jobs().get().execute.return_value = self.job_ready
            mock_client.jobs().run().execute.return_value = self.job_ready

            # Getting execution will raise error
            def raise_exception(*args, **kwargs):
                raise Exception("This is an intentional exception")

            monkeypatch.setattr("prefect_gcp.utilities.Execution.get", raise_exception)

            with pytest.raises(Exception):
                await cloud_run_worker.run(flow_run, cloud_run_worker_job_config)
            calls = list_mock_calls(mock_client, 2)
            for call, expected_call in zip(calls, expected_calls):
                assert call.startswith(expected_call)



from __future__ import annotations

from typing import (
    Any,
    Callable,
    TypeVar,
)

from typing_extensions import ParamSpec

from prefect import Flow
from prefect.flows import InfrastructureBoundFlow, bind_flow_to_infrastructure
from prefect_gcp.workers.cloud_run_v2 import CloudRunWorkerV2
from prefect_gcp.workers.vertex import VertexAIWorker

P = ParamSpec("P")
R = TypeVar("R")


def cloud_run(
    work_pool: str, **job_variables: Any
) -> Callable[[Flow[P, R]], InfrastructureBoundFlow[P, R]]:
    """
    Decorator that binds execution of a flow to a Cloud Run V2 work pool

    Args:
        work_pool: The name of the Cloud Run V2 work pool to use
        **job_variables: Additional job variables to use for infrastructure configuration

    Example:
        ```python
        from prefect import flow
        from prefect_gcp.experimental import cloud_run

        @cloud_run(work_pool="my-pool")
        @flow
        def my_flow():
            ...

        # This will run the flow in a Cloud Run job
        my_flow()
        ```
    """

    def decorator(flow: Flow[P, R]) -> InfrastructureBoundFlow[P, R]:
        return bind_flow_to_infrastructure(
            flow,
            work_pool=work_pool,
            job_variables=job_variables,
            worker_cls=CloudRunWorkerV2,
        )

    return decorator


def vertex_ai(
    work_pool: str, **job_variables: Any
) -> Callable[[Flow[P, R]], InfrastructureBoundFlow[P, R]]:
    """
    Decorator that binds execution of a flow to a Vertex AI work pool

    Args:
        work_pool: The name of the Vertex AI work pool to use
        **job_variables: Additional job variables to use for infrastructure configuration

    Example:
        ```python
        from prefect import flow
        from prefect_gcp.experimental import vertex_ai

        @vertex_ai(work_pool="my-pool")
        @flow
        def my_flow():
            ...

        # This will run the flow in a Vertex AI custom training job
        my_flow()
        ```
    """

    def decorator(flow: Flow[P, R]) -> InfrastructureBoundFlow[P, R]:
        return bind_flow_to_infrastructure(
            flow,
            work_pool=work_pool,
            job_variables=job_variables,
            worker_cls=VertexAIWorker,
        )

    return decorator



from __future__ import annotations

import logging
from pathlib import Path
from typing import Optional, TypedDict, cast

import typer

import prefect_gcp.credentials

logger = logging.getLogger("prefect_gcp.experimental.bundles.upload")


class UploadBundleToGcsOutput(TypedDict):
    """
    The output of the `upload_bundle_to_gcs` step.
    """

    bucket: str
    key: str


def upload_bundle_to_gcs(
    local_filepath: Path,
    bucket: str,
    key: str,
    gcp_credentials_block_name: str | None = None,
) -> UploadBundleToGcsOutput:
    """
    Uploads a bundle file to a GCS bucket.

    Args:
        local_filepath: The path to the bundle file to upload.
        bucket: The name of the GCS bucket to upload the bundle to.
        key: The key (path) to upload the bundle to in the GCS bucket.
        gcp_credentials_block_name: The name of the GCP credentials block to use.

    Returns:
        A dictionary containing the bucket and key of the uploaded bundle.
    """
    if not local_filepath.exists():
        raise ValueError(f"Bundle file not found: {local_filepath}")

    key = key or local_filepath.name

    if gcp_credentials_block_name:
        logger.debug(
            "Loading GCP credentials from block %s", gcp_credentials_block_name
        )
        gcp_credentials = cast(
            prefect_gcp.credentials.GcpCredentials,
            prefect_gcp.credentials.GcpCredentials.load(
                gcp_credentials_block_name,
                _sync=True,  # pyright: ignore[reportCallIssue] _sync is needed to prevent incidental async
            ),
        )
    else:
        logger.debug("Loading default GCP credentials")
        gcp_credentials = prefect_gcp.credentials.GcpCredentials()

    gcs_client = gcp_credentials.get_cloud_storage_client()

    try:
        logger.debug(
            "Uploading bundle from path %s to GCS bucket %s with key %s",
            local_filepath,
            bucket,
            key,
        )
        gcs_client.bucket(bucket).blob(key).upload_from_file(local_filepath)  # pyright: ignore[reportUnknownMemberType] Incomplete type hints
    except Exception as e:
        raise RuntimeError(f"Failed to upload bundle to GCS: {e}")

    return {"bucket": bucket, "key": key}


def _cli_wrapper(
    local_filepath: Path = typer.Argument(
        ..., help="The path to the bundle file to upload."
    ),
    bucket: str = typer.Option(
        ..., help="The name of the GCS bucket to upload the bundle to."
    ),
    key: str = typer.Option(
        ..., help="The key (path) to upload the bundle to in the GCS bucket."
    ),
    gcp_credentials_block_name: Optional[str] = typer.Option(
        None,
        help="The name of the GCP credentials block to use for authentication. If not provided, the default credentials will be used.",
    ),
) -> UploadBundleToGcsOutput:
    """
    Uploads a bundle file to a GCS bucket.
    """
    return upload_bundle_to_gcs(local_filepath, bucket, key, gcp_credentials_block_name)


if __name__ == "__main__":
    typer.run(_cli_wrapper)



from . import _version
from .bigquery import BigQueryWarehouse  # noqa
from .cloud_storage import GcsBucket  # noqa
from .credentials import GcpCredentials  # noqa
from .secret_manager import GcpSecret  # noqa
from .workers.vertex import VertexAIWorker  # noqa
from .workers.cloud_run import CloudRunWorker  # noqa
from .workers.cloud_run_v2 import CloudRunWorkerV2  # noqa


__version__ = _version.__version__



from pathlib import Path
from typing import Any
from unittest.mock import MagicMock

import pytest
from prefect_gcp.experimental.bundles.execute import execute_bundle_from_gcs
from pydantic_core import to_json

from prefect.runner import Runner


@pytest.fixture
def mock_bundle_data() -> dict[str, Any]:
    return {
        "context": "foo",
        "serialize_function": "bar",
        "flow_run": {"name": "buzz", "id": "123"},
    }


def test_execute_bundle_from_gcs_success(
    gcp_credentials: MagicMock,
    mock_bundle_data: dict[str, Any],
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    # Mock the GcpCredentials.load method
    monkeypatch.setattr(
        "prefect_gcp.credentials.GcpCredentials.load",
        MagicMock(return_value=gcp_credentials),
    )

    # Mock the Runner and its execute_bundle method
    mock_runner = MagicMock(spec=Runner)
    monkeypatch.setattr("prefect.runner.Runner", MagicMock(return_value=mock_runner))

    # Call the function
    bucket = "test-bucket"
    key = "test-key.json"

    def mock_download_to_file(path: str) -> None:
        Path(path).write_bytes(to_json(mock_bundle_data))

    # wow, i love mocking so much
    gcp_credentials.get_cloud_storage_client.return_value.bucket.return_value.blob.return_value.download_to_file.side_effect = mock_download_to_file

    execute_bundle_from_gcs(
        bucket=bucket,
        key=key,
        gcp_credentials_block_name="test-credentials",
    )

    # Verify the GCS client was called correctly
    gcs_client = gcp_credentials.get_cloud_storage_client()
    gcs_client.bucket.assert_called_once_with(bucket)
    gcs_client.bucket(bucket).blob.assert_called_once_with(key)
    gcs_client.bucket(bucket).blob(key).download_to_file.assert_called_once()

    # Verify the Runner was called correctly
    mock_runner.execute_bundle.assert_called_once_with(mock_bundle_data)


def test_execute_bundle_from_gcs_with_default_credentials(
    gcp_credentials: MagicMock,
    mock_bundle_data: dict[str, Any],
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    # Mock the GcpCredentials constructor
    monkeypatch.setattr(
        "prefect_gcp.credentials.GcpCredentials",
        MagicMock(return_value=gcp_credentials),
    )

    # Mock the Runner and its execute_bundle method
    mock_runner = MagicMock(spec=Runner)
    monkeypatch.setattr("prefect.runner.Runner", MagicMock(return_value=mock_runner))

    # Call the function
    bucket = "test-bucket"
    key = "test-key.json"

    def mock_download_to_file(path: str) -> None:
        Path(path).write_bytes(to_json(mock_bundle_data))

    # Set up the mock for download_to_file
    gcp_credentials.get_cloud_storage_client.return_value.bucket.return_value.blob.return_value.download_to_file.side_effect = mock_download_to_file

    execute_bundle_from_gcs(
        bucket=bucket,
        key=key,
    )

    # Verify the GCS client was called correctly
    gcs_client = gcp_credentials.get_cloud_storage_client()
    gcs_client.bucket.assert_called_once_with(bucket)
    gcs_client.bucket(bucket).blob.assert_called_once_with(key)
    gcs_client.bucket(bucket).blob(key).download_to_file.assert_called_once()

    # Verify the Runner was called correctly
    mock_runner.execute_bundle.assert_called_once_with(mock_bundle_data)


def test_execute_bundle_from_gcs_download_failure(
    gcp_credentials: MagicMock,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    # Mock the GcpCredentials.load method
    monkeypatch.setattr(
        "prefect_gcp.credentials.GcpCredentials.load",
        MagicMock(return_value=gcp_credentials),
    )

    # Call the function
    bucket = "test-bucket"
    key = "test-key.json"

    # Mock the GCS client to raise an exception
    gcs_client = gcp_credentials.get_cloud_storage_client()
    gcs_client.bucket(bucket).blob(key).download_to_file.side_effect = Exception(
        "Download failed"
    )

    # Call the function and expect it to raise a RuntimeError
    with pytest.raises(
        RuntimeError, match="Failed to download bundle from GCS: Download failed"
    ):
        execute_bundle_from_gcs(
            bucket=bucket,
            key=key,
            gcp_credentials_block_name="test-credentials",
        )


def test_execute_bundle_from_gcs_execution_failure(
    gcp_credentials: MagicMock,
    mock_bundle_data: dict[str, Any],
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    # Mock the GcpCredentials.load method
    monkeypatch.setattr(
        "prefect_gcp.credentials.GcpCredentials.load",
        MagicMock(return_value=gcp_credentials),
    )

    # Mock the Runner and its execute_bundle method to raise an exception
    mock_runner = MagicMock(spec=Runner)
    mock_runner.execute_bundle.side_effect = Exception("Execution failed")
    monkeypatch.setattr("prefect.runner.Runner", MagicMock(return_value=mock_runner))

    # Call the function
    bucket = "test-bucket"
    key = "test-key.json"

    def mock_download_to_file(path: str) -> None:
        Path(path).write_bytes(to_json(mock_bundle_data))

    # Set up the mock for download_to_file
    gcp_credentials.get_cloud_storage_client.return_value.bucket.return_value.blob.return_value.download_to_file.side_effect = mock_download_to_file

    # Call the function and expect it to raise a RuntimeError
    with pytest.raises(
        RuntimeError, match="Failed to download bundle from GCS: Execution failed"
    ):
        execute_bundle_from_gcs(
            bucket=bucket,
            key=key,
            gcp_credentials_block_name="test-credentials",
        )



from pathlib import Path
from unittest.mock import MagicMock

import pytest
from prefect_gcp.credentials import GcpCredentials
from prefect_gcp.experimental.bundles.upload import upload_bundle_to_gcs


@pytest.fixture
def mock_bundle_file(tmp_path: Path) -> Path:
    bundle_file = tmp_path / "test_bundle.tar.gz"
    bundle_file.write_text("test bundle content")
    return Path(bundle_file)


def test_upload_bundle_to_gcs_success(
    gcp_credentials: GcpCredentials,
    mock_bundle_file: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        "prefect_gcp.credentials.GcpCredentials.load",
        MagicMock(return_value=gcp_credentials),
    )
    bucket = "test-bucket"
    key = "test-key.json"

    result = upload_bundle_to_gcs(
        local_filepath=mock_bundle_file,
        bucket=bucket,
        key=key,
        gcp_credentials_block_name="test-credentials",
    )

    assert result == {"bucket": bucket, "key": key}

    # Verify the GCS client was called correctly
    gcs_client = gcp_credentials.get_cloud_storage_client()
    gcs_client.bucket.assert_called_once_with(bucket)
    gcs_client.bucket(bucket).blob.assert_called_once_with(key)
    gcs_client.bucket(bucket).blob(key).upload_from_file.assert_called_once_with(
        mock_bundle_file
    )


def test_upload_bundle_to_gcs_with_default_key(
    gcp_credentials: GcpCredentials,
    mock_bundle_file: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        "prefect_gcp.credentials.GcpCredentials.load",
        MagicMock(return_value=gcp_credentials),
    )
    bucket = "test-bucket"

    result = upload_bundle_to_gcs(
        local_filepath=mock_bundle_file,
        bucket=bucket,
        key="",
        gcp_credentials_block_name="test-credentials",
    )

    assert result == {"bucket": bucket, "key": mock_bundle_file.name}

    # Verify the GCS client was called with the default key
    gcs_client = gcp_credentials.get_cloud_storage_client()
    gcs_client.bucket(bucket).blob.assert_called_once_with(mock_bundle_file.name)


def test_upload_bundle_to_gcs_with_default_credentials(
    gcp_credentials: GcpCredentials,
    mock_bundle_file: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        "prefect_gcp.credentials.GcpCredentials",
        MagicMock(return_value=gcp_credentials),
    )
    bucket = "test-bucket"
    key = "test-key.tar.gz"

    result = upload_bundle_to_gcs(
        local_filepath=mock_bundle_file,
        bucket=bucket,
        key=key,
    )

    assert result == {"bucket": bucket, "key": key}

    # Verify the GCS client was called correctly
    gcs_client = gcp_credentials.get_cloud_storage_client()
    gcs_client.bucket(bucket).blob(key).upload_from_file.assert_called_once_with(
        mock_bundle_file
    )


def test_upload_bundle_to_gcs_file_not_found(
    gcp_credentials: GcpCredentials, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        "prefect_gcp.credentials.GcpCredentials.load",
        MagicMock(return_value=gcp_credentials),
    )
    non_existent_file = Path("/non/existent/file.tar.gz")
    bucket = "test-bucket"
    key = "test-key.tar.gz"

    with pytest.raises(ValueError, match=f"Bundle file not found: {non_existent_file}"):
        upload_bundle_to_gcs(
            local_filepath=non_existent_file,
            bucket=bucket,
            key=key,
            gcp_credentials_block_name="test-credentials",
        )


def test_upload_bundle_to_gcs_upload_failure(
    gcp_credentials: GcpCredentials,
    mock_bundle_file: Path,
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    monkeypatch.setattr(
        "prefect_gcp.credentials.GcpCredentials.load",
        MagicMock(return_value=gcp_credentials),
    )
    bucket = "test-bucket"
    key = "test-key.tar.gz"

    # Mock the upload to raise an exception
    gcs_client = gcp_credentials.get_cloud_storage_client()
    gcs_client.bucket(bucket).blob(key).upload_from_file.side_effect = Exception(
        "Upload failed"
    )

    with pytest.raises(
        RuntimeError, match="Failed to upload bundle to GCS: Upload failed"
    ):
        upload_bundle_to_gcs(
            local_filepath=mock_bundle_file,
            bucket=bucket,
            key=key,
            gcp_credentials_block_name="test-credentials",
        )



import uuid

import pytest
from prefect_gcp.credentials import GcpCredentials
from prefect_gcp.models.cloud_run_v2 import SecretKeySelector
from prefect_gcp.workers.cloud_run_v2 import CloudRunWorkerJobV2Configuration

from prefect.client.schemas.objects import FlowRun


@pytest.fixture
def job_body():
    return {
        "client": "prefect",
        "launchStage": None,
        "template": {
            "template": {
                "maxRetries": None,
                "timeout": None,
                "vpcAccess": {
                    "connector": None,
                },
                "containers": [
                    {
                        "env": [],
                        "command": None,
                        "args": "-m prefect.engine",
                        "resources": {
                            "limits": {
                                "cpu": None,
                                "memory": None,
                            },
                        },
                    },
                ],
            }
        },
    }


@pytest.fixture
def flow_run():
    return FlowRun(flow_id=uuid.uuid4(), name="my-flow-run-name")


@pytest.fixture
def cloud_run_worker_v2_job_config(service_account_info, job_body):
    return CloudRunWorkerJobV2Configuration(
        name="my-job-name",
        job_body=job_body,
        credentials=GcpCredentials(service_account_info=service_account_info),
        region="us-central1",
        timeout=86400,
        env={"ENV1": "VALUE1", "ENV2": "VALUE2"},
    )


class TestCloudRunWorkerJobV2ConfigurationFiltering:
    def test_populate_env_filters_plaintext_api_key_when_secret_configured(
        self, cloud_run_worker_v2_job_config
    ):
        # Add plaintext API key to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_KEY"] = "plaintext-api-key"
        cloud_run_worker_v2_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="prefect-api-key", version="latest"
        )
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        # Should not contain plaintext version
        assert {"name": "PREFECT_API_KEY", "value": "plaintext-api-key"} not in env_vars
        # Should contain secret version
        assert {
            "name": "PREFECT_API_KEY",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-api-key", "version": "latest"}
            },
        } in env_vars
        # Other env vars should still be present
        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars

    def test_populate_env_filters_plaintext_auth_string_when_secret_configured(
        self, cloud_run_worker_v2_job_config
    ):
        # Add plaintext auth string to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_AUTH_STRING"] = (
            "plaintext-auth-string"
        )
        cloud_run_worker_v2_job_config.prefect_api_auth_string_secret = (
            SecretKeySelector(secret="prefect-auth-string", version="latest")
        )
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        # Should not contain plaintext version
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "value": "plaintext-auth-string",
        } not in env_vars
        # Should contain secret version
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-auth-string", "version": "latest"}
            },
        } in env_vars
        # Other env vars should still be present
        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars

    def test_populate_env_filters_both_plaintext_when_secrets_configured(
        self, cloud_run_worker_v2_job_config
    ):
        # Add plaintext versions to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_KEY"] = "plaintext-api-key"
        cloud_run_worker_v2_job_config.env["PREFECT_API_AUTH_STRING"] = (
            "plaintext-auth-string"
        )
        cloud_run_worker_v2_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="prefect-api-key", version="latest"
        )
        cloud_run_worker_v2_job_config.prefect_api_auth_string_secret = (
            SecretKeySelector(secret="prefect-auth-string", version="latest")
        )
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        # Should not contain plaintext versions
        assert {"name": "PREFECT_API_KEY", "value": "plaintext-api-key"} not in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "value": "plaintext-auth-string",
        } not in env_vars
        # Should contain secret versions
        assert {
            "name": "PREFECT_API_KEY",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-api-key", "version": "latest"}
            },
        } in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-auth-string", "version": "latest"}
            },
        } in env_vars
        # Other env vars should still be present
        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars

    def test_populate_env_keeps_plaintext_when_no_secrets_configured(
        self, cloud_run_worker_v2_job_config
    ):
        # Add plaintext versions to env but don't configure secrets
        cloud_run_worker_v2_job_config.env["PREFECT_API_KEY"] = "plaintext-api-key"
        cloud_run_worker_v2_job_config.env["PREFECT_API_AUTH_STRING"] = (
            "plaintext-auth-string"
        )
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        # Should contain plaintext versions since no secrets are configured
        assert {"name": "PREFECT_API_KEY", "value": "plaintext-api-key"} in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "value": "plaintext-auth-string",
        } in env_vars
        # Other env vars should still be present
        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars

    def test_populate_env_filters_api_key_when_configured_via_env_from_secrets(
        self, cloud_run_worker_v2_job_config
    ):
        # Add plaintext API key to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_KEY"] = "plaintext-api-key"
        # Configure API key via env_from_secrets instead of dedicated field
        cloud_run_worker_v2_job_config.env_from_secrets = {
            "PREFECT_API_KEY": SecretKeySelector(
                secret="prefect-api-key", version="latest"
            )
        }
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        # Should not contain plaintext version
        assert {"name": "PREFECT_API_KEY", "value": "plaintext-api-key"} not in env_vars
        # Should contain secret version from env_from_secrets
        assert {
            "name": "PREFECT_API_KEY",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-api-key", "version": "latest"}
            },
        } in env_vars
        # Other env vars should still be present
        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars

    def test_populate_env_filters_auth_string_when_configured_via_env_from_secrets(
        self, cloud_run_worker_v2_job_config
    ):
        # Add plaintext auth string to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_AUTH_STRING"] = (
            "plaintext-auth-string"
        )
        # Configure auth string via env_from_secrets instead of dedicated field
        cloud_run_worker_v2_job_config.env_from_secrets = {
            "PREFECT_API_AUTH_STRING": SecretKeySelector(
                secret="prefect-auth-string", version="latest"
            )
        }
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        # Should not contain plaintext version
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "value": "plaintext-auth-string",
        } not in env_vars
        # Should contain secret version from env_from_secrets
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-auth-string", "version": "latest"}
            },
        } in env_vars
        # Other env vars should still be present
        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars

    def test_populate_env_filters_both_when_configured_via_env_from_secrets(
        self, cloud_run_worker_v2_job_config
    ):
        # Add plaintext versions to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_KEY"] = "plaintext-api-key"
        cloud_run_worker_v2_job_config.env["PREFECT_API_AUTH_STRING"] = (
            "plaintext-auth-string"
        )
        # Configure both via env_from_secrets
        cloud_run_worker_v2_job_config.env_from_secrets = {
            "PREFECT_API_KEY": SecretKeySelector(
                secret="prefect-api-key", version="latest"
            ),
            "PREFECT_API_AUTH_STRING": SecretKeySelector(
                secret="prefect-auth-string", version="latest"
            ),
        }
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        # Should not contain plaintext versions
        assert {"name": "PREFECT_API_KEY", "value": "plaintext-api-key"} not in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "value": "plaintext-auth-string",
        } not in env_vars
        # Should contain secret versions from env_from_secrets
        assert {
            "name": "PREFECT_API_KEY",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-api-key", "version": "latest"}
            },
        } in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "valueSource": {
                "secretKeyRef": {"secret": "prefect-auth-string", "version": "latest"}
            },
        } in env_vars
        # Other env vars should still be present
        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars

    def test_populate_env_prioritizes_dedicated_secret_fields_over_env_from_secrets(
        self, cloud_run_worker_v2_job_config
    ):
        # Add plaintext versions to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_KEY"] = "plaintext-api-key"
        cloud_run_worker_v2_job_config.env["PREFECT_API_AUTH_STRING"] = (
            "plaintext-auth-string"
        )
        # Configure via both dedicated fields and env_from_secrets
        cloud_run_worker_v2_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="dedicated-api-key", version="latest"
        )
        cloud_run_worker_v2_job_config.prefect_api_auth_string_secret = (
            SecretKeySelector(secret="dedicated-auth-string", version="latest")
        )
        cloud_run_worker_v2_job_config.env_from_secrets = {
            "PREFECT_API_KEY": SecretKeySelector(
                secret="env-from-secrets-api-key", version="latest"
            ),
            "PREFECT_API_AUTH_STRING": SecretKeySelector(
                secret="env-from-secrets-auth-string", version="latest"
            ),
        }
        cloud_run_worker_v2_job_config._populate_env()

        env_vars = cloud_run_worker_v2_job_config.job_body["template"]["template"][
            "containers"
        ][0]["env"]

        # Should not contain plaintext versions
        assert {"name": "PREFECT_API_KEY", "value": "plaintext-api-key"} not in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "value": "plaintext-auth-string",
        } not in env_vars

        # Should contain dedicated field secrets (should be added after env_from_secrets)
        assert {
            "name": "PREFECT_API_KEY",
            "valueSource": {
                "secretKeyRef": {"secret": "dedicated-api-key", "version": "latest"}
            },
        } in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "valueSource": {
                "secretKeyRef": {"secret": "dedicated-auth-string", "version": "latest"}
            },
        } in env_vars

        # Should also contain env_from_secrets versions
        assert {
            "name": "PREFECT_API_KEY",
            "valueSource": {
                "secretKeyRef": {
                    "secret": "env-from-secrets-api-key",
                    "version": "latest",
                }
            },
        } in env_vars
        assert {
            "name": "PREFECT_API_AUTH_STRING",
            "valueSource": {
                "secretKeyRef": {
                    "secret": "env-from-secrets-auth-string",
                    "version": "latest",
                }
            },
        } in env_vars

        # Other env vars should still be present
        assert {"name": "ENV1", "value": "VALUE1"} in env_vars
        assert {"name": "ENV2", "value": "VALUE2"} in env_vars


class TestCloudRunWorkerJobV2ConfigurationWarnings:
    def test_warn_about_plaintext_api_key(
        self, cloud_run_worker_v2_job_config, flow_run, caplog
    ):
        # Add plaintext API key to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_KEY"] = "plaintext-api-key"

        cloud_run_worker_v2_job_config.prepare_for_flow_run(
            flow_run=flow_run, deployment=None, flow=None
        )

        assert (
            "PREFECT_API_KEY is provided as a plaintext environment variable"
            in caplog.text
        )
        assert "consider providing it as a secret using" in caplog.text
        assert "'prefect_api_key_secret' or 'env_from_secrets'" in caplog.text

    def test_warn_about_plaintext_auth_string(
        self, cloud_run_worker_v2_job_config, flow_run, caplog
    ):
        # Add plaintext auth string to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_AUTH_STRING"] = (
            "plaintext-auth-string"
        )

        cloud_run_worker_v2_job_config.prepare_for_flow_run(
            flow_run=flow_run, deployment=None, flow=None
        )

        assert (
            "PREFECT_API_AUTH_STRING is provided as a plaintext environment variable"
            in caplog.text
        )
        assert "consider providing it as a secret using" in caplog.text
        assert "'prefect_api_auth_string_secret' or 'env_from_secrets'" in caplog.text

    def test_warn_about_both_plaintext_credentials(
        self, cloud_run_worker_v2_job_config, flow_run, caplog
    ):
        # Add both plaintext credentials to env
        cloud_run_worker_v2_job_config.env["PREFECT_API_KEY"] = "plaintext-api-key"
        cloud_run_worker_v2_job_config.env["PREFECT_API_AUTH_STRING"] = (
            "plaintext-auth-string"
        )

        cloud_run_worker_v2_job_config.prepare_for_flow_run(
            flow_run=flow_run, deployment=None, flow=None
        )

        # Should warn about both
        assert "PREFECT_API_KEY is provided as a plaintext" in caplog.text
        assert "PREFECT_API_AUTH_STRING is provided as a plaintext" in caplog.text

    def test_no_warning_when_api_key_secret_configured(
        self, cloud_run_worker_v2_job_config, flow_run, caplog
    ):
        # Add plaintext API key but configure secret
        cloud_run_worker_v2_job_config.env["PREFECT_API_KEY"] = "plaintext-api-key"
        cloud_run_worker_v2_job_config.prefect_api_key_secret = SecretKeySelector(
            secret="prefect-api-key", version="latest"
        )

        cloud_run_worker_v2_job_config.prepare_for_flow_run(
            flow_run=flow_run, deployment=None, flow=None
        )

        # Should not warn since secret is configured
        assert "PREFECT_API_KEY is provided as a plaintext" not in caplog.text

    def test_no_warning_when_auth_string_in_env_from_secrets(
        self, cloud_run_worker_v2_job_config, flow_run, caplog
    ):
        # Add plaintext auth string but configure it in env_from_secrets
        cloud_run_worker_v2_job_config.env["PREFECT_API_AUTH_STRING"] = (
            "plaintext-auth-string"
        )
        cloud_run_worker_v2_job_config.env_from_secrets = {
            "PREFECT_API_AUTH_STRING": SecretKeySelector(
                secret="prefect-auth-string", version="latest"
            )
        }

        cloud_run_worker_v2_job_config.prepare_for_flow_run(
            flow_run=flow_run, deployment=None, flow=None
        )

        # Should not warn since secret is configured via env_from_secrets
        assert "PREFECT_API_AUTH_STRING is provided as a plaintext" not in caplog.text

    def test_no_warning_when_no_plaintext_credentials(
        self, cloud_run_worker_v2_job_config, flow_run, caplog
    ):
        # Don't add any plaintext credentials
        cloud_run_worker_v2_job_config.prepare_for_flow_run(
            flow_run=flow_run, deployment=None, flow=None
        )

        # Should not warn since no plaintext credentials are present
        assert "is provided as a plaintext environment variable" not in caplog.text



import uuid
from unittest.mock import MagicMock

import pydantic
import pytest
from google.cloud.aiplatform_v1.types.custom_job import Scheduling
from google.cloud.aiplatform_v1.types.job_state import JobState
from prefect_gcp.workers.vertex import (
    VertexAIWorker,
    VertexAIWorkerJobConfiguration,
    VertexAIWorkerResult,
)

from prefect.client.schemas import FlowRun


@pytest.fixture
def job_config(service_account_info, gcp_credentials):
    return VertexAIWorkerJobConfiguration(
        name="my-custom-ai-job",
        region="ashenvale",
        credentials=gcp_credentials,
        job_spec={
            "service_account_name": "my-service-account",
            "maximum_run_time_hours": 1,
            "worker_pool_specs": [
                {
                    "replica_count": 1,
                    "container_spec": {
                        "image_uri": "gcr.io/your-project/your-repo:latest",
                        "command": ["python", "-m", "prefect.engine"],
                    },
                    "machine_spec": {
                        "machine_type": "n1-standard-4",
                        "accelerator_type": "NVIDIA_TESLA_K80",
                        "accelerator_count": 1,
                    },
                    "disk_spec": {
                        "boot_disk_type": "pd-ssd",
                        "boot_disk_size_gb": 100,
                    },
                }
            ],
            "scheduling": {
                "strategy": "FLEX_START",
                "max_wait_duration": "1800s",
            },
            "enable_web_access": True,
            "enable_dashboard_access": True,
        },
    )


@pytest.fixture
def flow_run():
    return FlowRun(flow_id=uuid.uuid4(), name="my-flow-run-name")


class TestVertexAIWorkerJobConfiguration:
    async def test_validate_empty_job_spec(self, gcp_credentials):
        base_job_template = VertexAIWorker.get_default_base_job_template()
        base_job_template["job_configuration"]["job_spec"] = {}
        base_job_template["job_configuration"]["region"] = "us-central1"

        with pytest.raises(pydantic.ValidationError) as excinfo:
            await VertexAIWorkerJobConfiguration.from_template_and_values(
                base_job_template, {"credentials": gcp_credentials}
            )

        assert len(errs := excinfo.value.errors()) == 1
        loc, msg, type_ = [errs[0].get(k) for k in ("loc", "msg", "type")]
        assert "job_spec" in str(loc)
        assert type_ == "value_error"
        assert "/maximum_run_time_hours" in msg
        assert "/worker_pool_specs" in msg

    async def test_validate_incomplete_worker_pool_spec(self, gcp_credentials):
        base_job_template = VertexAIWorker.get_default_base_job_template()
        base_job_template["job_configuration"]["job_spec"] = {
            "maximum_run_time_hours": 1,
            "worker_pool_specs": [
                {
                    "replica_count": 1,
                    "container_spec": {"command": ["some", "command"]},
                    "machine_spec": {
                        "accelerator_type": "NVIDIA_TESLA_K80",
                    },
                },
            ],
        }
        base_job_template["job_configuration"]["region"] = "us-central1"

        with pytest.raises(pydantic.ValidationError) as excinfo:
            await VertexAIWorkerJobConfiguration.from_template_and_values(
                base_job_template, {"credentials": gcp_credentials}
            )

        assert len(errs := excinfo.value.errors()) == 1
        loc, msg, type_ = [errs[0].get(k) for k in ("loc", "msg", "type")]
        assert "job_spec" in str(loc)
        assert type_ == "value_error"
        assert "/worker_pool_specs/0/container_spec/image_uri" in msg
        assert "/worker_pool_specs/0/disk_spec" in msg
        assert "/worker_pool_specs/0/machine_spec/machine_type" in msg

    def test_gcp_project(self, job_config: VertexAIWorkerJobConfiguration):
        assert job_config.project == "gcp_credentials_project"

    def test_job_name(self, flow_run, job_config: VertexAIWorkerJobConfiguration):
        job_config.prepare_for_flow_run(flow_run, None, None)
        assert job_config.job_name.startswith("my-custom-ai-job")

        job_config.name = None
        job_config.prepare_for_flow_run(flow_run, None, None)
        assert job_config.job_name.startswith("my-flow-run-name")

    async def test_missing_service_account(self, flow_run, job_config):
        job_config.job_spec["service_account_name"] = None
        job_config.credentials._service_account_email = None

        with pytest.raises(
            ValueError, match="A service account is required for the Vertex job"
        ):
            job_config.prepare_for_flow_run(flow_run, None, None)

    def test_valid_command_formatting(
        self, flow_run, job_config: VertexAIWorkerJobConfiguration
    ):
        job_config.prepare_for_flow_run(flow_run, None, None)
        assert ["python", "-m", "prefect.engine"] == job_config.job_spec[
            "worker_pool_specs"
        ][0]["container_spec"]["command"]

        job_config.job_spec["worker_pool_specs"][0]["container_spec"]["command"] = (
            "echo -n hello"
        )
        job_config.prepare_for_flow_run(flow_run, None, None)
        assert ["echo", "-n", "hello"] == job_config.job_spec["worker_pool_specs"][0][
            "container_spec"
        ]["command"]


class TestVertexAIWorker:
    async def test_successful_worker_run(self, flow_run, job_config):
        async with VertexAIWorker("test-pool") as worker:
            job_config.prepare_for_flow_run(flow_run, None, None)
            result = await worker.run(flow_run=flow_run, configuration=job_config)
            assert (
                job_config.credentials.job_service_async_client.create_custom_job.call_count
                == 1
            )
            custom_job_spec = job_config.credentials.job_service_async_client.create_custom_job.call_args[
                1
            ]["custom_job"].job_spec

            assert custom_job_spec.service_account == "my-service-account"

            assert (
                job_config.credentials.job_service_async_client.get_custom_job.call_count
                == 1
            )
            assert result == VertexAIWorkerResult(
                status_code=0, identifier="mock_display_name"
            )

    async def test_initiate_run_does_not_wait_for_completion(
        self, flow_run, job_config
    ):
        async with VertexAIWorker("test-pool") as worker:
            job_config.prepare_for_flow_run(flow_run, None, None)
            await worker._initiate_run(flow_run=flow_run, configuration=job_config)

            job_config.credentials.job_service_async_client.create_custom_job.assert_called_once()
            # The worker doesn't wait for the job to complete, so the get_custom_job call should not have been made
            job_config.credentials.job_service_async_client.get_custom_job.assert_not_called()

    async def test_missing_scheduling(self, flow_run, job_config):
        job_config.job_spec["scheduling"] = None

        async with VertexAIWorker("test-pool") as worker:
            job_config.prepare_for_flow_run(flow_run, None, None)
            await worker.run(flow_run=flow_run, configuration=job_config)

    async def test_params_worker_run(self, flow_run, job_config):
        async with VertexAIWorker("test-pool") as worker:
            # Initialize scheduling parameters
            maximum_run_time_hours = job_config.job_spec["maximum_run_time_hours"]
            max_wait_duration = job_config.job_spec["scheduling"]["max_wait_duration"]
            timeout = str(maximum_run_time_hours * 60 * 60) + "s"
            scheduling = Scheduling(
                timeout=timeout, max_wait_duration=max_wait_duration
            )

            # Additional params
            enable_web_access = job_config.job_spec["enable_web_access"]
            enable_dashboard_access = job_config.job_spec["enable_dashboard_access"]

            job_config.prepare_for_flow_run(flow_run, None, None)
            result = await worker.run(flow_run=flow_run, configuration=job_config)

            custom_job_spec = job_config.credentials.job_service_async_client.create_custom_job.call_args[
                1
            ]["custom_job"].job_spec

            # Assert scheduling parameters
            assert custom_job_spec.scheduling.timeout == scheduling.timeout
            assert (
                custom_job_spec.scheduling.strategy == Scheduling.Strategy["FLEX_START"]
            )
            assert (
                custom_job_spec.scheduling.max_wait_duration
                == scheduling.max_wait_duration
            )
            # Assert additional parameters
            assert custom_job_spec.enable_web_access == enable_web_access
            assert custom_job_spec.enable_dashboard_access == enable_dashboard_access

            assert (
                job_config.credentials.job_service_async_client.get_custom_job.call_count
                == 1
            )
            assert result == VertexAIWorkerResult(
                status_code=0, identifier="mock_display_name"
            )

    async def test_failed_worker_run(self, flow_run, job_config):
        job_config.prepare_for_flow_run(flow_run, None, None)
        error_msg = "something went kablooey"
        error_job_display_name = "catastrophization"
        job_config.credentials.job_service_async_client.get_custom_job.return_value = (
            MagicMock(
                name="error_mock_name",
                state=JobState.JOB_STATE_FAILED,
                error=MagicMock(message=error_msg),
                display_name=error_job_display_name,
            )
        )
        async with VertexAIWorker("test-pool") as worker:
            with pytest.raises(RuntimeError, match=error_msg):
                await worker.run(flow_run=flow_run, configuration=job_config)

            assert (
                job_config.credentials.job_service_async_client.create_custom_job.call_count
                == 1
            )
            assert (
                job_config.credentials.job_service_async_client.get_custom_job.call_count
                == 1
            )

    async def test_cancelled_worker_run(self, flow_run, job_config):
        job_config.prepare_for_flow_run(flow_run, None, None)
        job_display_name = "a-job-well-done"
        job_config.credentials.job_service_async_client.get_custom_job.return_value = (
            MagicMock(
                name="cancelled_mock_name",
                state=JobState.JOB_STATE_CANCELLED,
                error=MagicMock(message=""),
                display_name=job_display_name,
            )
        )
        async with VertexAIWorker("test-pool") as worker:
            result = await worker.run(flow_run=flow_run, configuration=job_config)
            assert (
                job_config.credentials.job_service_async_client.create_custom_job.call_count
                == 1
            )
            assert (
                job_config.credentials.job_service_async_client.get_custom_job.call_count
                == 1
            )
            assert result == VertexAIWorkerResult(
                status_code=1, identifier=job_display_name
            )



from unittest.mock import Mock

from pydantic import VERSION as PYDANTIC_VERSION

if PYDANTIC_VERSION.startswith("2."):
    pass
else:
    pass

import pytest
from prefect_gcp.utilities import Execution, Job

executions_return_value = {
    "metadata": {"name": "test-name", "namespace": "test-namespace"},
    "spec": {"MySpec": "spec"},
    "status": {"logUri": "test-log-uri"},
}

jobs_return_value = {
    "metadata": {"name": "Test", "namespace": "test-namespace"},
    "spec": {"MySpec": "spec"},
    "status": {
        "conditions": [{"type": "Ready", "dog": "cat"}],
        "latestCreatedExecution": {"puppy": "kitty"},
    },
}


class MockExecution(Mock):
    call_count = 0

    def __init__(self, succeeded=False, *args, **kwargs):
        super().__init__()
        self.log_uri = "test_uri"
        self._succeeded = succeeded

    def is_running(self):
        MockExecution.call_count += 1

        if self.call_count > 2:
            return False
        return True

    def condition_after_completion(self):
        return {"message": "test"}

    def succeeded(self):
        return self._succeeded

    @classmethod
    def get(cls, *args, **kwargs):
        return cls()


class TestJob:
    @pytest.mark.parametrize(
        "ready_condition,expected_value",
        [({"status": "True"}, True), ({"status": "False"}, False), ({}, False)],
    )
    def test_is_ready(self, ready_condition, expected_value):
        job = Job(
            metadata={},
            spec={},
            status={},
            name="test",
            ready_condition=ready_condition,
            execution_status={},
        )
        assert job.is_ready() == expected_value

    @pytest.mark.parametrize(
        "status,expected_value",
        [
            ({}, {}),
            ({"conditions": []}, {}),
            ({"conditions": [{"type": "Dog", "val": "value"}]}, {}),
            (
                {"conditions": [{"type": "Ready", "val": "value"}]},
                {"type": "Ready", "val": "value"},
            ),
            (
                {
                    "conditions": [
                        {"type": "Dog", "val": "value"},
                        {"type": "Ready", "val": "value"},
                    ]
                },
                {"type": "Ready", "val": "value"},
            ),
        ],
    )
    def test_get_ready_condition(self, status, expected_value):
        assert Job._get_ready_condition({"status": status}) == expected_value

    @pytest.mark.parametrize(
        "status,expected_value",
        [
            ({}, {}),
            ({"latestCreatedExecution": {}}, {}),
            ({"latestCreatedExecution": {"some": "val"}}, {"some": "val"}),
        ],
    )
    def test_get_execution_status(self, status, expected_value):
        assert Job._get_execution_status({"status": status}) == expected_value

    @pytest.mark.parametrize(
        "execution_status,expected_value",
        [
            ({}, True),  # Has no execution
            (
                {"completionTimestamp": None},
                True,
            ),  # Has an execution with no completion timestamp
            (
                {"completionTimestamp": "Exists"},
                False,
            ),  # Has an execution and it has a completion timestamp
        ],
    )
    def test_has_execution_in_progress(self, execution_status, expected_value):
        job = Job(
            metadata={},
            spec={},
            status={},
            name="test",
            ready_condition={},
            execution_status=execution_status,
        )
        assert job.has_execution_in_progress() == expected_value


class TestExecution:
    def test_succeeded_responds_true(self):
        """Desired behavior: `succeeded()` should return true if execution.status
        contains a list element that is a dict with a key "type" and a value of
        "Completed" and a key "status" and a value of "True"
        """
        execution = Execution(
            name="Test",
            namespace="test-namespace",
            metadata={},
            spec={},
            status={"conditions": [{"type": "Completed", "status": "True"}]},
            log_uri="",
        )
        assert execution.succeeded()

    @pytest.mark.parametrize(
        "conditions",
        [
            [],
            [{"type": "Dog", "status": "True"}],
            [{"type": "Completed", "status": "False"}],
            [{"type": "Completed", "status": "Dog"}],
        ],
    )
    def test_succeeded_responds_false(self, conditions):
        """Desired behavior: `succeeded()` should return False if execution.status
        lacks a list element that is a dict with a key "type" and a value of
        "Completed" and a key "status" and a value of "True".

        This could be a situation where there is no element containing the key, or
        the element with the key has a status that is not "True".
        """
        execution = Execution(
            name="Test",
            namespace="test-namespace",
            metadata={},
            spec={},
            status={"conditions": conditions},
            log_uri="",
        )
        assert not execution.succeeded()

    @pytest.mark.parametrize(
        "status,expected_value", [({}, True), ({"completionTime": "xyz"}, False)]
    )
    def test_is_running(self, status, expected_value):
        """Desired behavior: `is_running()` should return True if there if
        execution.status lack a key "completionTime", otherwise return False
        """
        execution = Execution(
            name="Test",
            namespace="test-namespace",
            metadata={},
            spec={},
            status=status,
            log_uri="",
        )
        assert execution.is_running() == expected_value

    @pytest.mark.parametrize(
        "conditions, expected_value",
        [
            ([], None),
            ([{"type": "Dog", "status": "True"}], None),
            (
                [{"type": "Completed", "status": "False"}],
                {"type": "Completed", "status": "False"},
            ),
            (
                [
                    {"type": "Dog", "status": "True"},
                    {"type": "Completed", "status": "False"},
                ],
                {"type": "Completed", "status": "False"},
            ),
        ],
    )
    def test_condition_after_completion_returns_correct_condition(
        self, conditions, expected_value
    ):
        """Desired behavior: `condition_after_completion()`
        should return the list element from execution.status that contains a dict
        with a key "type" and a value of "Completed" if it exists, else None
        """
        execution = Execution(
            name="Test",
            namespace="test-namespace",
            metadata={},
            spec={},
            status={"conditions": conditions},
            log_uri="",
        )
        assert execution.condition_after_completion() == expected_value
